# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""PKI CLI."""

import logging
import os
from pathlib import Path

from click import group
from click import option
from click import pass_context
from click import password_option
from click import Path as ClickPath

from openfl.component.ca.ca import CA_CONFIG_JSON
from openfl.component.ca.ca import CA_PASSWORD_FILE
from openfl.component.ca.ca import CA_PKI_DIR
from openfl.component.ca.ca import CA_STEP_CONFIG_DIR
from openfl.component.ca.ca import certify
from openfl.component.ca.ca import get_ca_bin_paths
from openfl.component.ca.ca import get_token
from openfl.component.ca.ca import install
from openfl.component.ca.ca import remove_ca
from openfl.component.ca.ca import run_ca

logger = logging.getLogger(__name__)

CA_URL = 'localhost:9123'


@group()
@pass_context
def pki(context):
    """Manage Step-ca PKI."""
    context.obj['group'] = 'pki'


@pki.command(name='run')
@option('-p', '--ca-path', required=True,
        help='The ca path', type=ClickPath())
def run(ca_path):
    """Run CA server."""
    ca_path = Path(ca_path).absolute()
    step_config_dir = ca_path / CA_STEP_CONFIG_DIR
    pki_dir = ca_path / CA_PKI_DIR
    password_file = pki_dir / CA_PASSWORD_FILE
    ca_json = step_config_dir / CA_CONFIG_JSON
    _, step_ca_path = get_ca_bin_paths(ca_path)
    if (not os.path.exists(step_config_dir) or not os.path.exists(pki_dir)
            or not os.path.exists(password_file) or not os.path.exists(ca_json)
            or not os.path.exists(step_ca_path)):
        logger.warning('CA is not installed or corrupted, please install it first')
        return
    run_ca(step_ca_path, password_file, ca_json)


@pki.command(name='install')
@option('-p', '--ca-path', required=True,
        help='The ca path', type=ClickPath())
@password_option(prompt='The password will encrypt some ca files \nEnter the password')
@option('--ca-url', required=False, default=CA_URL)
def install_(ca_path, password, ca_url):
    """Create a ca workspace."""
    ca_path = Path(ca_path).absolute()
    install(ca_path, ca_url, password)


@pki.command(name='uninstall')
@option('-p', '--ca-path', required=True,
        help='The CA path', type=ClickPath())
def uninstall(ca_path):
    """Remove step-CA."""
    ca_path = Path(ca_path).absolute()
    remove_ca(ca_path)


@pki.command(name='get-token')
@option('-n', '--name', required=True)
@option('--ca-url', required=False, default=CA_URL)
@option('-p', '--ca-path', default='.',
        help='The CA path', type=ClickPath(exists=True))
def get_token_(name, ca_url, ca_path):
    """
    Create authentication token.

    Args:
        name: common name for following certificate
                    (aggregator fqdn or collaborator name)
        ca_url: full url of CA server
        ca_path: the path to CA binaries
    """
    ca_path = Path(ca_path).absolute()
    token = get_token(name, ca_url, ca_path)
    print('Token:')
    print(token)


@pki.command(name='certify')
@option('-n', '--name', required=True)
@option('-t', '--token', 'token_with_cert', required=True)
@option('-c', '--certs-path', required=False, default=Path('.') / 'cert',
        help='The path where certificates will be stored', type=ClickPath())
@option('-p', '--ca-path', default='.', help='The path to CA client',
        type=ClickPath(exists=True), required=False)
def certify_(name, token_with_cert, certs_path, ca_path):
    """Create an envoy workspace."""
    certs_path = Path(certs_path).absolute()
    ca_path = Path(ca_path).absolute()
    certs_path.mkdir(parents=True, exist_ok=True)
    certify(name, certs_path, token_with_cert, ca_path)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import tensorflow.keras as ke
from tensorflow.keras import Sequential
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Flatten

from openfl.federated import KerasTaskRunner


class KerasCNN(KerasTaskRunner):
    """A basic convolutional neural network model."""

    def __init__(self, **kwargs):
        """
        Initialize.

        Args:
            **kwargs: Additional parameters to pass to the function
        """
        super().__init__(**kwargs)

        self.model = self.build_model(self.feature_shape, self.data_loader.num_classes, **kwargs)

        self.initialize_tensorkeys_for_functions()

        self.model.summary(print_fn=self.logger.info)

        if self.data_loader is not None:
            self.logger.info(f'Train Set Size : {self.get_train_data_size()}')
            self.logger.info(f'Valid Set Size : {self.get_valid_data_size()}')

    def build_model(self,
                    input_shape,
                    num_classes,
                    conv_kernel_size=(4, 4),
                    conv_strides=(2, 2),
                    conv1_channels_out=16,
                    conv2_channels_out=32,
                    final_dense_inputsize=100,
                    **kwargs):
        """
        Define the model architecture.

        Args:
            input_shape (numpy.ndarray): The shape of the data
            num_classes (int): The number of classes of the dataset

        Returns:
            tensorflow.python.keras.engine.sequential.Sequential: The model defined in Keras

        """
        model = Sequential()

        model.add(Conv2D(conv1_channels_out,
                         kernel_size=conv_kernel_size,
                         strides=conv_strides,
                         activation='relu',
                         input_shape=input_shape))

        model.add(Conv2D(conv2_channels_out,
                         kernel_size=conv_kernel_size,
                         strides=conv_strides,
                         activation='relu'))

        model.add(Flatten())

        model.add(Dense(final_dense_inputsize, activation='relu'))

        model.add(Dense(num_classes, activation='softmax'))

        model.compile(loss=ke.losses.categorical_crossentropy,
                      optimizer=ke.optimizers.Adam(),
                      metrics=['accuracy'])

        # initialize the optimizer variables
        opt_vars = model.optimizer.variables()

        for v in opt_vars:
            v.initializer.run(session=self.sess)

        return model



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Static grouped assigner module."""

from .assigner import Assigner


class StaticGroupedAssigner(Assigner):
    r"""
    The task assigner maintains a list of tasks.

    Also it decides the policy for
    which collaborator should run those tasks
    There may be many types of policies implemented, but a natural place to
    start is with a:

    StaticGroupedAssigner  - Given a set of task groups, and a list of
                             collaborators for that group, assign tasks for
                             of collaborators in the federation. After assigning
                             the tasks to collaborator, those tasks
                             should be carried out each round (no reassignment
                             between rounds)
    GroupedAssigner -        Given task groups and a list of collaborators that
                             belong to that task group, carry out tasks for
                             each round of experiment

    Args:
        task_groups* (list of obj): task groups to assign.

    Note:
        \* - Plan setting.
    """

    def __init__(self, task_groups, **kwargs):
        """Initialize."""
        self.task_groups = task_groups
        super().__init__(**kwargs)

    def define_task_assignments(self):
        """All of the logic to set up the map of tasks to collaborators is done here."""
        cols_amount = sum([
            len(group['collaborators']) for group in self.task_groups
        ])
        authorized_cols_amount = len(self.authorized_cols)

        unique_cols = {
            col
            for group in self.task_groups
            for col in group['collaborators']
        }
        unique_authorized_cols = set(self.authorized_cols)

        assert (cols_amount == authorized_cols_amount and unique_cols == unique_authorized_cols), (
            f'Collaborators in each group must be distinct: '
            f'{unique_cols}, {unique_authorized_cols}'
        )

        # Start by finding all of the tasks in all specified groups
        self.all_tasks_in_groups = list({
            task
            for group in self.task_groups
            for task in group['tasks']
        })

        # Initialize the map of collaborators for a given task on a given round
        for task in self.all_tasks_in_groups:
            self.collaborators_for_task[task] = {
                i: [] for i in range(self.rounds)
            }

        for group in self.task_groups:
            group_col_list = group['collaborators']
            self.task_group_collaborators[group['name']] = group_col_list
            for col in group_col_list:
                # For now, we assume that collaborators have the same tasks for
                # every round
                self.collaborator_tasks[col] = {
                    i: group['tasks'] for i in range(self.rounds)
                }
            # Now populate reverse lookup of tasks->group
            for task in group['tasks']:
                for round_ in range(self.rounds):
                    # This should append the list of collaborators performing
                    # that task
                    self.collaborators_for_task[task][round_] += group_col_list

    def get_tasks_for_collaborator(self, collaborator_name, round_number):
        """Get tasks for the collaborator specified."""
        return self.collaborator_tasks[collaborator_name][round_number]

    def get_collaborators_for_task(self, task_name, round_number):
        """Get collaborators for the task specified."""
        return self.collaborators_for_task[task_name][round_number]



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

from logging import getLogger

import numpy as np
from torch.utils.tensorboard import SummaryWriter
from torchvision import datasets
from torchvision import transforms

logger = getLogger(__name__)

writer = None


def get_writer():
    """Create global writer object."""
    global writer
    if not writer:
        writer = SummaryWriter('./logs/cnn_mnist', flush_secs=5)


def write_metric(node_name, task_name, metric_name, metric, round_number):
    """Write metric callback."""
    get_writer()
    writer.add_scalar(f'{node_name}/{task_name}/{metric_name}', metric, round_number)


def one_hot(labels, classes):
    """
    One Hot encode a vector.

    Args:
        labels (list):  List of labels to onehot encode
        classes (int): Total number of categorical classes

    Returns:
        np.array: Matrix of one-hot encoded labels
    """
    return np.eye(classes)[labels]


def _load_raw_datashards(shard_num, collaborator_count, transform=None):
    """
    Load the raw data by shard.

    Returns tuples of the dataset shard divided into training and validation.

    Args:
        shard_num (int): The shard number to use
        collaborator_count (int): The number of collaborators in the federation
        transform: torchvision.transforms.Transform to apply to images

    Returns:
        2 tuples: (image, label) of the training, validation dataset
    """
    train_data, val_data = (
        datasets.MNIST('data', train=train, download=True, transform=transform)
        for train in (True, False)
    )
    X_train_tot, y_train_tot = train_data.train_data, train_data.train_labels
    X_valid_tot, y_valid_tot = val_data.test_data, val_data.test_labels

    # create the shards
    shard_num = int(shard_num)
    X_train = X_train_tot[shard_num::collaborator_count].unsqueeze(1).float()
    y_train = y_train_tot[shard_num::collaborator_count]

    X_valid = X_valid_tot[shard_num::collaborator_count].unsqueeze(1).float()
    y_valid = y_valid_tot[shard_num::collaborator_count]

    return (X_train, y_train), (X_valid, y_valid)


def load_mnist_shard(shard_num, collaborator_count,
                     categorical=False, channels_last=True, **kwargs):
    """
    Load the MNIST dataset.

    Args:
        shard_num (int): The shard to use from the dataset
        collaborator_count (int): The number of collaborators in the
                                  federation
        categorical (bool): True = convert the labels to one-hot encoded
                            vectors (Default = True)
        channels_last (bool): True = The input images have the channels
                              last (Default = True)
        **kwargs: Additional parameters to pass to the function

    Returns:
        list: The input shape
        int: The number of classes
        numpy.ndarray: The training data
        numpy.ndarray: The training labels
        numpy.ndarray: The validation data
        numpy.ndarray: The validation labels
    """
    num_classes = 10

    (X_train, y_train), (X_valid, y_valid) = _load_raw_datashards(
        shard_num, collaborator_count, transform=transforms.ToTensor())

    logger.info(f'MNIST > X_train Shape : {X_train.shape}')
    logger.info(f'MNIST > y_train Shape : {y_train.shape}')
    logger.info(f'MNIST > Train Samples : {X_train.shape[0]}')
    logger.info(f'MNIST > Valid Samples : {X_valid.shape[0]}')

    if categorical:
        # convert class vectors to binary class matrices
        y_train = one_hot(y_train, num_classes)
        y_valid = one_hot(y_valid, num_classes)

    return num_classes, X_train, y_train, X_valid, y_valid



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""TinyImageNet Shard Descriptor."""

import glob
import logging
import os
import shutil
from pathlib import Path
from typing import Tuple

from PIL import Image

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor

logger = logging.getLogger(__name__)


class TinyImageNetDataset(ShardDataset):
    """TinyImageNet shard dataset class."""

    NUM_IMAGES_PER_CLASS = 500

    def __init__(self, data_folder: Path, data_type='train', rank=1, worldsize=1):
        """Initialize TinyImageNetDataset."""
        self.data_type = data_type
        self._common_data_folder = data_folder
        self._data_folder = os.path.join(data_folder, data_type)
        self.labels = {}  # fname - label number mapping
        self.image_paths = sorted(
            glob.iglob(
                os.path.join(self._data_folder, '**', '*.JPEG'),
                recursive=True
            )
        )[rank - 1::worldsize]
        with open(os.path.join(self._common_data_folder, 'wnids.txt'), 'r') as fp:
            self.label_texts = sorted([text.strip() for text in fp.readlines()])
        self.label_text_to_number = {text: i for i, text in enumerate(self.label_texts)}
        self.fill_labels()

    def __len__(self) -> int:
        """Return the len of the shard dataset."""
        return len(self.image_paths)

    def __getitem__(self, index: int) -> Tuple['Image', int]:
        """Return an item by the index."""
        file_path = self.image_paths[index]
        label = self.labels[os.path.basename(file_path)]
        return self.read_image(file_path), label

    def read_image(self, path: Path) -> Image:
        """Read the image."""
        img = Image.open(path)
        return img

    def fill_labels(self) -> None:
        """Fill labels."""
        if self.data_type == 'train':
            for label_text, i in self.label_text_to_number.items():
                for cnt in range(self.NUM_IMAGES_PER_CLASS):
                    self.labels[f'{label_text}_{cnt}.JPEG'] = i
        elif self.data_type == 'val':
            with open(os.path.join(self._data_folder, 'val_annotations.txt'), 'r') as fp:
                for line in fp.readlines():
                    terms = line.split('\t')
                    file_name, label_text = terms[0], terms[1]
                    self.labels[file_name] = self.label_text_to_number[label_text]


class TinyImageNetShardDescriptor(ShardDescriptor):
    """Shard descriptor class."""

    def __init__(
            self,
            data_folder: str = 'data',
            rank_worldsize: str = '1,1',
            **kwargs
    ):
        """Initialize TinyImageNetShardDescriptor."""
        self.common_data_folder = Path.cwd() / data_folder
        self.data_folder = Path.cwd() / data_folder / 'tiny-imagenet-200'
        self.download_data()
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))

    def download_data(self):
        """Download prepared shard dataset."""
        zip_file_path = self.common_data_folder / 'tiny-imagenet-200.zip'
        os.makedirs(self.common_data_folder, exist_ok=True)
        os.system(f'wget --no-clobber http://cs231n.stanford.edu/tiny-imagenet-200.zip'
                  f' -O {zip_file_path}')
        shutil.unpack_archive(str(zip_file_path), str(self.common_data_folder))

    def get_dataset(self, dataset_type):
        """Return a shard dataset by type."""
        return TinyImageNetDataset(
            data_folder=self.data_folder,
            data_type=dataset_type,
            rank=self.rank,
            worldsize=self.worldsize
        )

    @property
    def sample_shape(self):
        """Return the sample shape info."""
        return ['64', '64', '3']

    @property
    def target_shape(self):
        """Return the target shape info."""
        return ['64', '64']

    @property
    def dataset_description(self) -> str:
        """Return the shard dataset description."""
        return (f'TinyImageNetDataset dataset, shard number {self.rank}'
                f' out of {self.worldsize}')



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""openfl Native functions module.

This file defines openfl entrypoints to be used directly through python (not CLI)
"""

import logging
import os
from copy import copy
from logging import getLogger
from pathlib import Path

import flatten_json

import openfl.interface.aggregator as aggregator
import openfl.interface.collaborator as collaborator
import openfl.interface.workspace as workspace
from openfl.federated import Plan
from openfl.protocols import utils
from openfl.utilities import add_log_level
from openfl.utilities import split_tensor_dict_for_holdouts

logger = getLogger(__name__)

WORKSPACE_PREFIX = os.path.join(os.path.expanduser('~'), '.local', 'workspace')


def setup_plan(log_level='CRITICAL'):
    """
    Dump the plan with all defaults + overrides set.

    Args:
        save : bool (default=True)
            Whether to save the plan to disk

    Returns:
        plan : Plan object
    """
    plan_config = 'plan/plan.yaml'
    cols_config = 'plan/cols.yaml'
    data_config = 'plan/data.yaml'

    current_level = logging.root.level
    getLogger().setLevel(log_level)
    plan = Plan.parse(plan_config_path=Path(plan_config),
                      cols_config_path=Path(cols_config),
                      data_config_path=Path(data_config),
                      resolve=False)
    getLogger().setLevel(current_level)

    return plan


def flatten(config, return_complete=False):
    """Flatten nested config."""
    flattened_config = flatten_json.flatten(config, '.')
    if not return_complete:
        keys_to_remove = [
            k for k, v in flattened_config.items()
            if ('defaults' in k or v is None)]
    else:
        keys_to_remove = [k for k, v in flattened_config.items() if v is None]
    for k in keys_to_remove:
        del flattened_config[k]

    return flattened_config


def update_plan(override_config, plan=None, resolve=True):
    """
    Update the plan with the provided override and save it to disk.

    For a list of available override options, call `fx.get_plan()`

    Args:
        override_config : dict {"COMPONENT.settings.variable" : value or list of values}

    Returns:
        None
    """
    if plan is None:
        plan = setup_plan()
    flat_plan_config = flatten(plan.config, return_complete=True)

    org_list_keys_with_count = {}
    for k in flat_plan_config:
        k_split = k.rsplit('.', 1)
        if k_split[1].isnumeric():
            if k_split[0] in org_list_keys_with_count:
                org_list_keys_with_count[k_split[0]] += 1
            else:
                org_list_keys_with_count[k_split[0]] = 1

    for key, val in override_config.items():
        if key in org_list_keys_with_count:
            # remove old list corresponding to this key entirely
            for idx in range(org_list_keys_with_count[key]):
                del flat_plan_config[f'{key}.{idx}']
            logger.info(f'Updating {key} to {val}... ')
        elif key in flat_plan_config:
            logger.info(f'Updating {key} to {val}... ')
        else:
            # TODO: We probably need to validate the new key somehow
            logger.info(f'Did not find {key} in config. Make sure it should exist. Creating...')
        if type(val) == list:
            for idx, v in enumerate(val):
                flat_plan_config[f'{key}.{idx}'] = v
        else:
            flat_plan_config[key] = val

    plan.config = unflatten(flat_plan_config, '.')
    if resolve:
        plan.resolve()
    return plan


def unflatten(config, separator='.'):
    """Unfold `config` settings that have `separator` in their names."""
    config = flatten_json.unflatten_list(config, separator)
    return config


def setup_logging(level='INFO', log_file=None):
    """Initialize logging settings."""
    # Setup logging
    from logging import basicConfig
    from rich.console import Console
    from rich.logging import RichHandler
    import pkgutil
    if True if pkgutil.find_loader('tensorflow') else False:
        import tensorflow as tf
        tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
    metric = 25
    add_log_level('METRIC', metric)

    if isinstance(level, str):
        level = level.upper()

    handlers = []
    if log_file:
        fh = logging.FileHandler(log_file)
        formatter = logging.Formatter(
            '%(asctime)s %(levelname)s %(message)s %(filename)s:%(lineno)d'
        )
        fh.setFormatter(formatter)
        handlers.append(fh)

    console = Console(width=160)
    handlers.append(RichHandler(console=console))
    basicConfig(level=level, format='%(message)s',
                datefmt='[%X]', handlers=handlers)


def init(workspace_template: str = 'default', log_level: str = 'INFO',
         log_file: str = None, agg_fqdn: str = None, col_names=None):
    """
    Initialize the openfl package.

    It performs the following tasks:

         1. Creates a workspace in ~/.local/workspace (Equivalent to `fx
         workspace create --prefix ~/.local/workspace --template
         $workspace_template)
         2. Setup certificate authority (equivalent to `fx workspace certify`)
         3. Setup aggregator PKI (equivalent to `fx aggregator
         generate-cert-request` followed by `fx aggregator certify`)
         4. Setup list of collaborators (col_names) and their PKI. (Equivalent
         to running `fx collaborator generate-cert-request` followed by `fx
         collaborator certify` for each of the collaborators in col_names)
         5. Setup logging

    Args:
        workspace_template : str (default='default')
            The template that should be used as the basis for the experiment.
            Other options include are any of the template names [
            keras_cnn_mnist, tf_2dunet, tf_cnn_histology, mtorch_cnn_histology,
            torch_cnn_mnist]
        log_level : str
            Log level for logging. METRIC level is available
        log_file : str
            Name of the file in which the log will be duplicated
        agg_fqdn : str
           The local node's fully qualified domain name (if it can't be
           resolved automatically)
        col_names: list[str]
           The names of the collaborators that will be created. These
           collaborators will be set up to participate in the experiment, but
           are not required to

    Returns:
        None
    """
    if col_names is None:
        col_names = ['one', 'two']
    workspace.create(WORKSPACE_PREFIX, workspace_template)
    os.chdir(WORKSPACE_PREFIX)
    workspace.certify()
    aggregator.generate_cert_request(agg_fqdn)
    aggregator.certify(agg_fqdn, silent=True)
    data_path = 1
    for col_name in col_names:
        collaborator.generate_cert_request(
            col_name, str(data_path), silent=True, skip_package=True)
        collaborator.certify(col_name, silent=True)
        data_path += 1

    setup_logging(level=log_level, log_file=log_file)


def create_collaborator(plan, name, model, aggregator, nn):
    """
    Create the collaborator.

    Using the same plan object to create multiple collaborators leads to
    identical collaborator objects. This function can be removed once
    collaborator generation is fixed in openfl/federated/plan/plan.py
    """
    plan = copy(plan)

    return plan.get_collaborator(name, task_runner=model, client=aggregator, nn=nn)


def run_experiment(collaborator_dict: dict, override_config: dict = None, nn: bool = False):
    """
    Core function that executes the FL Plan.

    Args:
        collaborator_dict : dict {collaborator_name(str): FederatedModel}
            This dictionary defines which collaborators will participate in the
            experiment, as well as a reference to that collaborator's
            federated model.
        override_config : dict {flplan.key : flplan.value}
            Override any of the plan parameters at runtime using this
            dictionary. To get a list of the available options, execute
            `fx.get_plan()`

    Returns:
        final_federated_model : FederatedModel
            The final model resulting from the federated learning experiment
    """
    from sys import path

    if override_config is None:
        override_config = {}

    file = Path(__file__).resolve()
    root = file.parent.resolve()  # interface root, containing command modules
    work = Path.cwd().resolve()

    path.append(str(root))
    path.insert(0, str(work))

    # Update the plan if necessary
    plan = update_plan(override_config)
    # Overwrite plan values
    plan.authorized_cols = list(collaborator_dict)
    tensor_pipe = plan.get_tensor_pipe()

    # This must be set to the final index of the list (this is the last
    # tensorflow session to get created)
    plan.runner_ = list(collaborator_dict.values())[-1]
    model = plan.runner_

    # Initialize model weights
    init_state_path = plan.config['aggregator']['settings']['init_state_path']
    rounds_to_train = plan.config['aggregator']['settings']['rounds_to_train']
    if nn:
        tensor_dict, holdout_params = split_tensor_dict_for_holdouts(
            logger,
            model.get_tensor_dict(False)
        )

        model_snap = utils.construct_model_proto(tensor_dict=tensor_dict,
                                                 round_number=0,
                                                 tensor_pipe=tensor_pipe)

        logger.info(f'Creating Initial Weights File    🠆 {init_state_path}')

        utils.dump_proto(model_proto=model_snap, fpath=init_state_path)

    logger.info('Starting Experiment...')

    aggregator = plan.get_aggregator()

    # Create the collaborators
    collaborators = {
        collaborator: create_collaborator(
            plan, collaborator, collaborator_dict[collaborator], aggregator, nn
        ) for collaborator in plan.authorized_cols
    }

    for _ in range(rounds_to_train):
        for col in plan.authorized_cols:
            collaborator = collaborators[col]
            collaborator.run_simulation()

    # Set the weights for the final model
    if nn:
        model.rebuild_model(
            rounds_to_train - 1, aggregator.last_tensor_dict, validation=True)
    return model


def get_plan(fl_plan=None, indent=4, sort_keys=True):
    """Get string representation of current Plan."""
    import json
    if fl_plan is None:
        plan = setup_plan()
    else:
        plan = fl_plan
    flat_plan_config = flatten(plan.config)
    return json.dumps(flat_plan_config, indent=indent, sort_keys=sort_keys)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Envoy CLI."""

import logging
import shutil
import sys
from importlib import import_module
from pathlib import Path

import click
from click import group
from click import option
from click import pass_context
from click import Path as ClickPath
from dynaconf import Validator

from openfl.component.envoy.envoy import Envoy
from openfl.interface.cli_helper import WORKSPACE
from openfl.utilities import click_types
from openfl.utilities import merge_configs
from openfl.utilities.path_check import is_directory_traversal

logger = logging.getLogger(__name__)


@group()
@pass_context
def envoy(context):
    """Manage Federated Learning Envoy."""
    context.obj['group'] = 'envoy'


@envoy.command(name='start')
@option('-n', '--shard-name', required=True,
        help='Current shard name')
@option('-dh', '--director-host', required=True,
        help='The FQDN of the federation director', type=click_types.FQDN)
@option('-dp', '--director-port', required=True,
        help='The federation director port', type=click.IntRange(1, 65535))
@option('--tls/--disable-tls', default=True,
        is_flag=True, help='Use TLS or not (By default TLS is enabled)')
@option('-ec', '--envoy-config-path', default='envoy_config.yaml',
        help='The envoy config path', type=ClickPath(exists=True))
@option('-rc', '--root-cert-path', 'root_certificate', default=None,
        help='Path to a root CA cert', type=ClickPath(exists=True))
@option('-pk', '--private-key-path', 'private_key', default=None,
        help='Path to a private key', type=ClickPath(exists=True))
@option('-oc', '--public-cert-path', 'certificate', default=None,
        help='Path to a signed certificate', type=ClickPath(exists=True))
def start_(shard_name, director_host, director_port, tls, envoy_config_path,
           root_certificate, private_key, certificate):
    """Start the Envoy."""
    logger.info('🧿 Starting the Envoy.')
    if is_directory_traversal(envoy_config_path):
        click.echo('The shard config path is out of the openfl workspace scope.')
        sys.exit(1)

    config = merge_configs(
        settings_files=envoy_config_path,
        overwrite_dict={
            'root_certificate': root_certificate,
            'private_key': private_key,
            'certificate': certificate,
        },
        validators=[
            Validator('shard_descriptor.template', required=True),
            Validator('params.cuda_devices', default=[]),
        ],
    )

    if config.root_certificate:
        config.root_certificate = Path(config.root_certificate).absolute()
    if config.private_key:
        config.private_key = Path(config.private_key).absolute()
    if config.certificate:
        config.certificate = Path(config.certificate).absolute()

    # Parse envoy parameters
    envoy_params = config.get('params', {})

    # Build optional plugin components
    optional_plugins_section = config.get('optional_plugin_components')
    if optional_plugins_section is not None:
        for plugin_name, plugin_settings in optional_plugins_section.items():
            template = plugin_settings.get('template')
            if not template:
                raise Exception('You should put a template'
                                f'for plugin {plugin_name}')
            module_path, _, class_name = template.rpartition('.')
            plugin_params = plugin_settings.get('params', {})

            module = import_module(module_path)
            instance = getattr(module, class_name)(**plugin_params)
            envoy_params[plugin_name] = instance

    # Instantiate Shard Descriptor
    shard_descriptor = shard_descriptor_from_config(config.get('shard_descriptor', {}))
    envoy = Envoy(
        shard_name=shard_name,
        director_host=director_host,
        director_port=director_port,
        tls=tls,
        shard_descriptor=shard_descriptor,
        root_certificate=config.root_certificate,
        private_key=config.private_key,
        certificate=config.certificate,
        **envoy_params
    )

    envoy.start()


@envoy.command(name='create-workspace')
@option('-p', '--envoy-path', required=True,
        help='The Envoy path', type=ClickPath())
def create(envoy_path):
    """Create an envoy workspace."""
    if is_directory_traversal(envoy_path):
        click.echo('The Envoy path is out of the openfl workspace scope.')
        sys.exit(1)
    envoy_path = Path(envoy_path).absolute()
    if envoy_path.exists():
        if not click.confirm('Envoy workspace already exists. Recreate?',
                             default=True):
            sys.exit(1)
        shutil.rmtree(envoy_path)
    (envoy_path / 'cert').mkdir(parents=True, exist_ok=True)
    (envoy_path / 'logs').mkdir(parents=True, exist_ok=True)
    (envoy_path / 'data').mkdir(parents=True, exist_ok=True)
    shutil.copyfile(WORKSPACE / 'default/envoy_config.yaml',
                    envoy_path / 'envoy_config.yaml')
    shutil.copyfile(WORKSPACE / 'default/shard_descriptor.py',
                    envoy_path / 'shard_descriptor.py')
    shutil.copyfile(WORKSPACE / 'default/requirements.txt',
                    envoy_path / 'requirements.txt')


def shard_descriptor_from_config(shard_config: dict):
    """Build a shard descriptor from config."""
    template = shard_config.get('template')
    if not template:
        raise Exception('You should define a shard '
                        'descriptor template in the envoy config')
    class_name = template.split('.')[-1]
    module_path = '.'.join(template.split('.')[:-1])
    params = shard_config.get('params', {})

    module = import_module(module_path)
    instance = getattr(module, class_name)(**params)

    return instance



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Cryptography participant utilities."""
from typing import Tuple

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.x509.base import CertificateSigningRequest
from cryptography.x509.oid import NameOID


def generate_csr(common_name: str,
                 server: bool = False) -> Tuple[RSAPrivateKey, CertificateSigningRequest]:
    """Issue certificate signing request for server and client."""
    # Generate private key
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=3072,
        backend=default_backend()
    )

    builder = x509.CertificateSigningRequestBuilder()
    subject = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, common_name),
    ])
    builder = builder.subject_name(subject)
    builder = builder.add_extension(
        x509.BasicConstraints(ca=False, path_length=None), critical=True,
    )
    if server:
        builder = builder.add_extension(
            x509.ExtendedKeyUsage([x509.ExtendedKeyUsageOID.SERVER_AUTH]),
            critical=True
        )

    else:
        builder = builder.add_extension(
            x509.ExtendedKeyUsage([x509.ExtendedKeyUsageOID.CLIENT_AUTH]),
            critical=True
        )

    builder = builder.add_extension(
        x509.KeyUsage(
            digital_signature=True,
            key_encipherment=True,
            data_encipherment=False,
            key_agreement=False,
            content_commitment=False,
            key_cert_sign=False,
            crl_sign=False,
            encipher_only=False,
            decipher_only=False
        ),
        critical=True
    )

    builder = builder.add_extension(
        x509.SubjectAlternativeName([x509.DNSName(common_name)]),
        critical=False
    )

    # Sign the CSR
    csr = builder.sign(
        private_key=private_key, algorithm=hashes.SHA384(),
        backend=default_backend()
    )

    return private_key, csr



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Straggler handling module."""

from abc import ABC
from abc import abstractmethod


class StragglerHandlingFunction(ABC):
    """Federated Learning straggler handling interface."""

    @abstractmethod
    def straggler_cutoff_check(self, **kwargs):
        """
        Determines whether it is time to end the round early.
        Returns:
            bool
        """
        raise NotImplementedError



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Utilities."""

import os
from contextlib import contextmanager

import _pickle as p
import numpy as np
import torch
from torch.utils.data import Dataset


def to_device(obj, device, non_blocking=False):
    """Copy to device."""
    if isinstance(obj, torch.Tensor):
        return obj.to(device, non_blocking=non_blocking)

    if isinstance(obj, dict):
        return {k: to_device(v, device, non_blocking=non_blocking)
                for k, v in obj.items()}

    if isinstance(obj, list):
        return [to_device(v, device, non_blocking=non_blocking)
                for v in obj]

    if isinstance(obj, tuple):
        return tuple([to_device(v, device, non_blocking=non_blocking)
                     for v in obj])


@contextmanager
def task(_):
    """Yield."""
    yield


class DictionaryConcatDataset(Dataset):
    """Concate dictionaries."""

    def __init__(self, d_of_datasets):
        """Initialize."""
        self.d_of_datasets = d_of_datasets
        lengths = [len(d) for d in d_of_datasets.values()]
        self._length = min(lengths)
        self.keys = self.d_of_datasets.keys()
        assert min(lengths) == max(lengths), 'Length of the datasets should be the same'

    def __getitem__(self, idx):
        """Get item."""
        return {
            key: self.d_of_datasets[key][idx]
            for key in self.keys
        }

    def __len__(self):
        """Get length."""
        return self._length


def crop_chw(image, i, j, k, s=1):
    """Crop func."""
    if s == 1:
        h, w = i, j
    else:
        h = s * i
        w = s * j
    return image[:, h: h + k, w: w + k]


def cnn_output_size(h, k, s=1, p=0) -> int:
    """Output size.

    :param int H: input_size
    :param int K: filter_size
    :param int S: stride
    :param int P: padding
    :return:.

    """
    return 1 + (h - k + 2 * p) // s


def crop_image_chw(image, coord, k):
    """Crop func."""
    h, w = coord
    return image[:, h: h + k, w: w + k]


def load_binary(fpath, encoding='ASCII'):
    """Load binaries."""
    with open(fpath, 'rb') as f:
        return p.load(f, encoding=encoding)


def save_binary(d, fpath):
    """Save binary."""
    with open(fpath, 'wb') as f:
        p.dump(d, f)


def makedirpath(fpath: str):
    """Make path."""
    dpath = os.path.dirname(fpath)
    if dpath:
        os.makedirs(dpath, exist_ok=True)


def distribute_scores(score_masks, output_shape, k: int, s: int) -> np.ndarray:
    """Distribute scores."""
    n_all = score_masks.shape[0]
    results = [distribute_score(score_masks[n], output_shape, k, s) for n in range(n_all)]
    return np.asarray(results)


def distribute_score(score_mask, output_shape, k: int, s: int) -> np.ndarray:
    """Distribute scores."""
    h, w = output_shape
    mask = np.zeros([h, w], dtype=np.float32)
    cnt = np.zeros([h, w], dtype=np.int32)

    i, j = score_mask.shape[:2]
    for i_ in range(i):
        for j_ in range(j):
            h_, w_ = i_ * s, j_ * s

            mask[h_: h_ + k, w_: w_ + k] += score_mask[i_, j_]
            cnt[h_: h_ + k, w_: w_ + k] += 1

    cnt[cnt == 0] = 1

    return mask / cnt



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Director clients module."""

import logging
from datetime import datetime
from typing import List
from typing import Type

import grpc

from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor
from openfl.pipelines import NoCompressionPipeline, GenericPipeline
from openfl.protocols import director_pb2
from openfl.protocols import director_pb2_grpc
from openfl.protocols import interceptors
from openfl.protocols.utils import construct_model_proto
from openfl.protocols.utils import deconstruct_model_proto
from openfl.transport.grpc.director_server import CLIENT_ID_DEFAULT

from .grpc_channel_options import channel_options

logger = logging.getLogger(__name__)


class ShardDirectorClient:
    """The internal director client class."""

    def __init__(self, *, director_host, director_port, shard_name, tls=True,
                 root_certificate=None, private_key=None, certificate=None) -> None:
        """Initialize a shard director client object."""
        self.shard_name = shard_name
        director_addr = f'{director_host}:{director_port}'
        if not tls:
            channel = grpc.insecure_channel(director_addr, options=channel_options)
        else:
            if not (root_certificate and private_key and certificate):
                raise Exception('No certificates provided')
            try:
                with open(root_certificate, 'rb') as f:
                    root_certificate_b = f.read()
                with open(private_key, 'rb') as f:
                    private_key_b = f.read()
                with open(certificate, 'rb') as f:
                    certificate_b = f.read()
            except FileNotFoundError as exc:
                raise Exception(f'Provided certificate file is not exist: {exc.filename}')

            credentials = grpc.ssl_channel_credentials(
                root_certificates=root_certificate_b,
                private_key=private_key_b,
                certificate_chain=certificate_b
            )
            channel = grpc.secure_channel(director_addr, credentials, options=channel_options)
        self.stub = director_pb2_grpc.DirectorStub(channel)

    def report_shard_info(self, shard_descriptor: Type[ShardDescriptor],
                          cuda_devices: tuple) -> bool:
        """Report shard info to the director."""
        logger.info('Send report UpdateShardInfo')
        # True considered as successful registration
        shard_info = director_pb2.ShardInfo(
            shard_description=shard_descriptor.dataset_description,
            sample_shape=shard_descriptor.sample_shape,
            target_shape=shard_descriptor.target_shape
        )

        shard_info.node_info.name = self.shard_name
        shard_info.node_info.cuda_devices.extend(
            director_pb2.CudaDeviceInfo(index=cuda_device)
            for cuda_device in cuda_devices
        )

        request = director_pb2.UpdateShardInfoRequest(shard_info=shard_info)
        acknowledgement = self.stub.UpdateShardInfo(request)
        return acknowledgement.accepted

    def wait_experiment(self):
        """Wait an experiment data from the director."""
        logger.info('Send WaitExperiment request')
        response_iter = self.stub.WaitExperiment(self._get_experiment_data())
        logger.info('WaitExperiment response has received')
        response = next(response_iter)
        experiment_name = response.experiment_name
        if not experiment_name:
            raise Exception('No experiment')

        return experiment_name

    def get_experiment_data(self, experiment_name):
        """Get an experiment data from the director."""
        logger.info(f'Request experiment {experiment_name}')
        request = director_pb2.GetExperimentDataRequest(
            experiment_name=experiment_name,
            collaborator_name=self.shard_name
        )
        data_stream = self.stub.GetExperimentData(request)

        return data_stream

    def set_experiment_failed(
            self,
            experiment_name: str,
            error_code: int = 1,
            error_description: str = ''
    ):
        """Set the experiment failed."""
        request = director_pb2.SetExperimentFailedRequest(
            experiment_name=experiment_name,
            collaborator_name=self.shard_name,
            error_code=error_code,
            error_description=error_description
        )
        self.stub.SetExperimentFailed(request)

    def _get_experiment_data(self):
        """Generate the experiment data request."""
        yield director_pb2.WaitExperimentRequest(collaborator_name=self.shard_name)

    def send_health_check(
            self, *,
            envoy_name: str,
            is_experiment_running: bool,
            cuda_devices_info: List[dict] = None,
    ) -> int:
        """Send envoy health check."""
        status = director_pb2.UpdateEnvoyStatusRequest(
            name=envoy_name,
            is_experiment_running=is_experiment_running,
        )

        cuda_messages = []
        if cuda_devices_info is not None:
            try:
                cuda_messages = [
                    director_pb2.CudaDeviceInfo(**item)
                    for item in cuda_devices_info
                ]
            except Exception as e:
                logger.info(f'{e}')

        status.cuda_devices.extend(cuda_messages)

        logger.debug(f'Sending health check status: {status}')

        response = self.stub.UpdateEnvoyStatus(status)
        health_check_period = response.health_check_period.seconds

        return health_check_period


class DirectorClient:
    """Director client class for users."""

    def __init__(
            self, *,
            client_id: str,
            director_host: str,
            director_port: int,
            tls: bool,
            root_certificate: str,
            private_key: str,
            certificate: str,
    ) -> None:
        """Initialize director client object."""
        director_addr = f'{director_host}:{director_port}'
        if not tls:
            if not client_id:
                client_id = CLIENT_ID_DEFAULT
            channel = grpc.insecure_channel(director_addr, options=channel_options)
            headers = {
                'client_id': client_id,
            }
            header_interceptor = interceptors.headers_adder(headers)
            channel = grpc.intercept_channel(channel, header_interceptor)
        else:
            if not (root_certificate and private_key and certificate):
                raise Exception('No certificates provided')
            try:
                with open(root_certificate, 'rb') as f:
                    root_certificate_b = f.read()
                with open(private_key, 'rb') as f:
                    private_key_b = f.read()
                with open(certificate, 'rb') as f:
                    certificate_b = f.read()
            except FileNotFoundError as exc:
                raise Exception(f'Provided certificate file is not exist: {exc.filename}')

            credentials = grpc.ssl_channel_credentials(
                root_certificates=root_certificate_b,
                private_key=private_key_b,
                certificate_chain=certificate_b
            )

            channel = grpc.secure_channel(director_addr, credentials, options=channel_options)
        self.stub = director_pb2_grpc.DirectorStub(channel)

    def set_new_experiment(self, name, col_names, arch_path,
                           initial_tensor_dict=None, nn=True):
        """Send the new experiment to director to launch."""
        logger.info('SetNewExperiment')
        if initial_tensor_dict:
            model_proto = construct_model_proto(initial_tensor_dict, 0, GenericPipeline(nn))
            experiment_info_gen = self._get_experiment_info(
                arch_path=arch_path,
                name=name,
                col_names=col_names,
                model_proto=model_proto,
            )
            resp = self.stub.SetNewExperiment(experiment_info_gen)
            return resp

    def _get_experiment_info(self, arch_path, name, col_names, model_proto):
        with open(arch_path, 'rb') as arch:
            max_buffer_size = 2 * 1024 * 1024
            chunk = arch.read(max_buffer_size)
            while chunk != b'':
                if not chunk:
                    raise StopIteration
                # TODO: add hash or/and size to check
                experiment_info = director_pb2.ExperimentInfo(
                    name=name,
                    collaborator_names=col_names,
                    model_proto=model_proto
                )
                experiment_info.experiment_data.size = len(chunk)
                experiment_info.experiment_data.npbytes = chunk
                yield experiment_info
                chunk = arch.read(max_buffer_size)

    def get_dataset_info(self):
        """Request the dataset info from the director."""
        resp = self.stub.GetDatasetInfo(director_pb2.GetDatasetInfoRequest())
        return resp.shard_info.sample_shape, resp.shard_info.target_shape

    def _get_trained_model(self, experiment_name, model_type):
        """Get trained model RPC."""
        get_model_request = director_pb2.GetTrainedModelRequest(
            experiment_name=experiment_name,
            model_type=model_type,
        )
        model_proto_response = self.stub.GetTrainedModel(get_model_request)
        tensor_dict, _ = deconstruct_model_proto(
            model_proto_response.model_proto,
            NoCompressionPipeline(),
        )
        return tensor_dict

    def get_best_model(self, experiment_name):
        """Get best model method."""
        model_type = director_pb2.GetTrainedModelRequest.BEST_MODEL
        return self._get_trained_model(experiment_name, model_type)

    def get_last_model(self, experiment_name):
        """Get last model method."""
        model_type = director_pb2.GetTrainedModelRequest.LAST_MODEL
        return self._get_trained_model(experiment_name, model_type)

    def stream_metrics(self, experiment_name):
        """Stream metrics RPC."""
        request = director_pb2.GetMetricStreamRequest(experiment_name=experiment_name)
        for metric_message in self.stub.GetMetricStream(request):
            yield {
                'metric_origin': metric_message.metric_origin,
                'task_name': metric_message.task_name,
                'metric_name': metric_message.metric_name,
                'metric_value': metric_message.metric_value,
                'round': metric_message.round,
            }

    def remove_experiment_data(self, name):
        """Remove experiment data RPC."""
        request = director_pb2.RemoveExperimentRequest(experiment_name=name)
        response = self.stub.RemoveExperimentData(request)
        return response.acknowledgement

    def get_envoys(self, raw_result=False):
        """Get envoys info."""
        envoys = self.stub.GetEnvoys(director_pb2.GetEnvoysRequest())
        if raw_result:
            return envoys
        now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        result = {}
        for envoy in envoys.envoy_infos:
            result[envoy.shard_info.node_info.name] = {
                'shard_info': envoy.shard_info,
                'is_online': envoy.is_online or False,
                'is_experiment_running': envoy.is_experiment_running or False,
                'last_updated': datetime.fromtimestamp(
                    envoy.last_updated.seconds).strftime('%Y-%m-%d %H:%M:%S'),
                'current_time': now,
                'valid_duration': envoy.valid_duration,
                'experiment_name': 'ExperimentName Mock',
            }
        return result

    def get_experiments_list(self):
        """Get experiments list."""
        response = self.stub.GetExperimentsList(
            director_pb2.GetExperimentsListRequest()
        )
        return response.experiments

    def get_experiment_description(self, name):
        """Get experiment info."""
        response = self.stub.GetExperimentDescription(
            director_pb2.GetExperimentDescriptionRequest(name=name)
        )
        return response.experiment



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Aggregator module."""

import base64
import json
import os
import platform
import shutil
import signal
import subprocess
import time
import urllib.request
from logging import getLogger
from pathlib import Path
from subprocess import call

import requests
from click import confirm

logger = getLogger(__name__)

TOKEN_DELIMITER = '.'
CA_STEP_CONFIG_DIR = Path('step_config')
CA_PKI_DIR = Path('cert')
CA_PASSWORD_FILE = Path('pass_file')
CA_CONFIG_JSON = Path('config/ca.json')


def get_system_and_architecture():
    """Get system and architecture of machine."""
    uname_res = platform.uname()
    system = uname_res.system.lower()

    architecture_aliases = {
        'x86_64': 'amd64',
        'armv6l': 'armv6',
        'armv7l': 'armv7',
        'aarch64': 'arm64'
    }
    architecture = uname_res.machine.lower()
    for alias in architecture_aliases:
        if architecture == alias:
            architecture = architecture_aliases[alias]
            break

    return system, architecture


def download_step_bin(url, grep_name, architecture, prefix='.', confirmation=True):
    """
    Donwload step binaries from github.

    Args:
        url: address of latest release
        grep_name: name to grep over github assets
        architecture: architecture type to grep
        prefix: folder path to download
        confirmation: request user confirmation or not
    """
    if confirmation:
        confirm('CA binaries from github will be downloaded now', default=True, abort=True)
    result = requests.get(url)
    if result.status_code != 200:
        logger.warning('Can\'t download binaries from github. Please try lately.')
        return

    assets = result.json().get('assets', [])
    archive_urls = [
        a['browser_download_url']
        for a in assets
        if (grep_name in a['name'] and architecture in a['name']
            and 'application/gzip' in a['content_type'])
    ]
    if len(archive_urls) == 0:
        raise Exception('Applicable CA binaries from github were not found '
                        f'(name: {grep_name}, architecture: {architecture})')
    archive_url = archive_urls[-1]
    archive_url = archive_url.replace('https', 'http')
    name = archive_url.split('/')[-1]
    logger.info(f'Downloading {name}')
    urllib.request.urlretrieve(archive_url, f'{prefix}/{name}')
    shutil.unpack_archive(f'{prefix}/{name}', f'{prefix}/step')


def get_token(name, ca_url, ca_path='.'):
    """
    Create authentication token.

    Args:
        name: common name for following certificate
                    (aggregator fqdn or collaborator name)
        ca_url: full url of CA server
        ca_path: path to ca folder
    """
    ca_path = Path(ca_path)
    step_config_dir = ca_path / CA_STEP_CONFIG_DIR
    pki_dir = ca_path / CA_PKI_DIR
    step_path, _ = get_ca_bin_paths(ca_path)
    if not step_path:
        raise Exception('Step-CA is not installed!\nRun `fx pki install` first')

    priv_json = step_config_dir / 'secrets' / 'priv.json'
    pass_file = pki_dir / CA_PASSWORD_FILE
    root_crt = step_config_dir / 'certs' / 'root_ca.crt'
    try:
        token = subprocess.check_output(
            f'{step_path} ca token {name} '
            f'--key {priv_json} --root {root_crt} '
            f'--password-file {pass_file} 'f'--ca-url {ca_url}', shell=True)
    except subprocess.CalledProcessError as exc:
        logger.error(f'Error code {exc.returncode}: {exc.output}')
        return

    token = token.strip()
    token_b64 = base64.b64encode(token)

    with open(root_crt, mode='rb') as file:
        root_certificate_b = file.read()
    root_ca_b64 = base64.b64encode(root_certificate_b)

    return TOKEN_DELIMITER.join([
        token_b64.decode('utf-8'),
        root_ca_b64.decode('utf-8'),
    ])


def get_ca_bin_paths(ca_path):
    """Get paths of step binaries."""
    ca_path = Path(ca_path)
    step = None
    step_ca = None
    if (ca_path / 'step').exists():
        dirs = os.listdir(ca_path / 'step')
        for dir_ in dirs:
            if 'step_' in dir_:
                step = ca_path / 'step' / dir_ / 'bin' / 'step'
            if 'step-ca' in dir_:
                step_ca = ca_path / 'step' / dir_ / 'bin' / 'step-ca'
    return step, step_ca


def certify(name, cert_path: Path, token_with_cert, ca_path: Path):
    """Create an envoy workspace."""
    os.makedirs(cert_path, exist_ok=True)

    token, root_certificate = token_with_cert.split(TOKEN_DELIMITER)
    token = base64.b64decode(token).decode('utf-8')
    root_certificate = base64.b64decode(root_certificate)

    step_path, _ = get_ca_bin_paths(ca_path)
    if not step_path:
        url = 'https://api.github.com/repos/smallstep/cli/releases/latest'
        system, arch = get_system_and_architecture()
        download_step_bin(url, f'step_{system}', arch, prefix=ca_path)
        step_path, _ = get_ca_bin_paths(ca_path)
    if not step_path:
        raise Exception('Step-CA is not installed!\nRun `fx pki install` first')

    with open(f'{cert_path}/root_ca.crt', mode='wb') as file:
        file.write(root_certificate)
    call(f'{step_path} ca certificate {name} {cert_path}/{name}.crt '
         f'{cert_path}/{name}.key --kty EC --curve P-384 -f --token {token}', shell=True)


def remove_ca(ca_path):
    """Kill step-ca process and rm ca directory."""
    _check_kill_process('step-ca')
    shutil.rmtree(ca_path, ignore_errors=True)


def install(ca_path, ca_url, password):
    """
    Create certificate authority for federation.

    Args:
        ca_path: path to ca directory
        ca_url: url for ca server like: 'host:port'
        password: Simple password for encrypting root private keys

    """
    logger.info('Creating CA')

    ca_path = Path(ca_path)
    ca_path.mkdir(parents=True, exist_ok=True)
    step_config_dir = ca_path / CA_STEP_CONFIG_DIR
    os.environ['STEPPATH'] = str(step_config_dir)
    step_path, step_ca_path = get_ca_bin_paths(ca_path)

    if not (step_path and step_ca_path and step_path.exists() and step_ca_path.exists()):
        confirm('CA binaries from github will be downloaded now', default=True, abort=True)
        system, arch = get_system_and_architecture()
        url = 'https://api.github.com/repos/smallstep/certificates/releases/latest'
        download_step_bin(url, f'step-ca_{system}', arch, prefix=ca_path, confirmation=False)
        url = 'https://api.github.com/repos/smallstep/cli/releases/latest'
        download_step_bin(url, f'step_{system}', arch, prefix=ca_path, confirmation=False)
    step_config_dir = ca_path / CA_STEP_CONFIG_DIR
    if (not step_config_dir.exists()
            or confirm('CA exists, do you want to recreate it?', default=True)):
        _create_ca(ca_path, ca_url, password)
    _configure(step_config_dir)


def run_ca(step_ca, pass_file, ca_json):
    """Run CA server."""
    if _check_kill_process('step-ca', confirmation=True):
        logger.info('Up CA server')
        call(f'{step_ca} --password-file {pass_file} {ca_json}', shell=True)


def _check_kill_process(pstring, confirmation=False):
    """Kill process by name."""
    pids = []
    proc = subprocess.Popen(f'ps ax | grep {pstring} | grep -v grep',
                            shell=True, stdout=subprocess.PIPE)
    text = proc.communicate()[0].decode('utf-8')

    for line in text.splitlines():
        fields = line.split()
        pids.append(fields[0])

    if len(pids):
        if confirmation and not confirm('CA server is already running. Stop him?', default=True):
            return False
        for pid in pids:
            os.kill(int(pid), signal.SIGKILL)
        time.sleep(2)
    return True


def _create_ca(ca_path: Path, ca_url: str, password: str):
    """Create a ca workspace."""
    import os
    pki_dir = ca_path / CA_PKI_DIR
    step_config_dir = ca_path / CA_STEP_CONFIG_DIR

    pki_dir.mkdir(parents=True, exist_ok=True)
    step_config_dir.mkdir(parents=True, exist_ok=True)

    with open(f'{pki_dir}/pass_file', 'w') as f:
        f.write(password)
    os.chmod(f'{pki_dir}/pass_file', 0o600)
    step_path, step_ca_path = get_ca_bin_paths(ca_path)
    assert (step_path and step_ca_path and step_path.exists() and step_ca_path.exists())

    logger.info('Create CA Config')
    os.environ['STEPPATH'] = str(step_config_dir)
    shutil.rmtree(step_config_dir, ignore_errors=True)
    name = ca_url.split(':')[0]
    call(f'{step_path} ca init --name name --dns {name} '
         f'--address {ca_url}  --provisioner prov '
         f'--password-file {pki_dir}/pass_file', shell=True)

    call(f'{step_path} ca provisioner remove prov --all', shell=True)
    call(f'{step_path} crypto jwk create {step_config_dir}/certs/pub.json '
         f'{step_config_dir}/secrets/priv.json --password-file={pki_dir}/pass_file', shell=True)
    call(
        f'{step_path} ca provisioner add provisioner {step_config_dir}/certs/pub.json',
        shell=True
    )


def _configure(step_config_dir):
    conf_file = step_config_dir / CA_CONFIG_JSON
    with open(conf_file, 'r+') as f:
        data = json.load(f)
        data.setdefault('authority', {}).setdefault('claims', {})
        data['authority']['claims']['maxTLSCertDuration'] = f'{365 * 24}h'
        data['authority']['claims']['defaultTLSCertDuration'] = f'{365 * 24}h'
        data['authority']['claims']['maxUserSSHCertDuration'] = '24h'
        data['authority']['claims']['defaultUserSSHCertDuration'] = '24h'
        f.seek(0)
        json.dump(data, f, indent=4)
        f.truncate()



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""MXNet Framework Adapter plugin."""

from pickle import dumps
from pickle import loads
from typing import Dict

import mxnet as mx
import numpy as np
from mxnet import nd

from openfl.plugins.frameworks_adapters.framework_adapter_interface import (
    FrameworkAdapterPluginInterface
)


class FrameworkAdapterPlugin(FrameworkAdapterPluginInterface):
    """Framework adapter plugin class."""

    def __init__(self) -> None:
        """Initialize framework adapter."""

    @staticmethod
    def get_tensor_dict(model, optimizer=None) -> Dict[str, np.ndarray]:
        """
        Extract tensor dict from a model and an optimizer.

        Returns:
        dict {weight name: numpy ndarray}
        """
        state = {}
        if optimizer is not None:
            state = _get_optimizer_state(optimizer)

        model_params = model.collect_params()

        for param_name, param_tensor in model_params.items():
            if isinstance(param_tensor.data(), mx.ndarray.ndarray.NDArray):
                state[param_name] = param_tensor.list_data()[0].asnumpy()

        return state

    @staticmethod
    def set_tensor_dict(model, tensor_dict: Dict[str, np.ndarray],
                        optimizer=None, device=None) -> None:
        """
        Set tensor dict from a model and an optimizer.

        Given a dict {weight name: numpy ndarray} sets weights to
        the model and optimizer objects inplace.
        """
        if device is not None:
            device = mx.cpu() if device.startswith('cpu') else (
                mx.gpu(int(device.split(':')[1].strip()))
            )

        if optimizer is not None:
            _set_optimizer_state(optimizer, device, tensor_dict)
        model.collect_params().reset_ctx(device)

        model_params = model.collect_params()

        for param_name in model_params:
            model_params[param_name].set_data(nd.array(tensor_dict.pop(param_name), ctx=device))


def _get_optimizer_state(optimizer):
    """Return the optimizer state.

    Args:
        optimizer
    """
    states = loads(optimizer._updaters[0].get_states(dump_optimizer=False))
    result_states = {}
    for state_key, state_tuple in states.items():
        for state_ind, state in enumerate(state_tuple):
            result_states[f'opt_state__{state_key}__{state_ind}'] = state.asnumpy()

    return result_states


def _set_optimizer_state(optimizer, device, opt_state_dict):
    """Set the optimizer state.

    Args:
        optimizer:
        device:

    """
    state_keys, max_numstates = set(), 0
    for key in opt_state_dict.keys():
        if not key.startswith('opt_state'):
            continue
        _, part1, part2 = key.split('__')
        state_keys.add(int(part1))
        max_numstates = max(max_numstates, int(part2))

    out_state = {}
    for _ in range(len(state_keys)):
        key = state_keys.pop()
        state_vals = []
        for i in range(max_numstates + 1):
            state_vals.append(nd.array(opt_state_dict.pop(f'opt_state__{key}__{i}'), ctx=device))
        out_state[key] = tuple(state_vals)

    optimizer._updaters[0].set_states(dumps(out_state))



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Data package."""

import pkgutil
from warnings import catch_warnings
from warnings import simplefilter

with catch_warnings():
    simplefilter(action='ignore', category=FutureWarning)
    if pkgutil.find_loader('tensorflow'):
        # ignore deprecation warnings in command-line interface
        import tensorflow  # NOQA

from .loader import DataLoader  # NOQA

if pkgutil.find_loader('tensorflow'):
    from .loader_tf import TensorFlowDataLoader  # NOQA
    from .loader_keras import KerasDataLoader  # NOQA
    from .federated_data import FederatedDataSet  # NOQA

if pkgutil.find_loader('torch'):
    from .loader_pt import PyTorchDataLoader  # NOQA
    from .federated_data import FederatedDataSet  # NOQA
if pkgutil.find_loader('torch') and pkgutil.find_loader('tensorflow'):
    from .loader_fe import FastEstimatorDataLoader  # NOQA



"""Copyright (C) 2020-2021 Intel Corporation
   SPDX-License-Identifier: Apache-2.0

Licensed subject to the terms of the separately executed evaluation
license agreement between Intel Corporation and you.
"""
from logging import getLogger
from os import getcwd
from os import path
from os import remove
from typing import Dict
from typing import Tuple
from zipfile import ZipFile

import numpy as np
import requests

logger = getLogger(__name__)


def download_data_() -> str:
    """Download data.

    Returns:
      string: relative path to data file
    """
    pkg = 'fra-eng.zip'   # Language file: change this to change the language
    data_dir = 'data'
    url = 'https://www.manythings.org/anki/' + pkg
    filename = pkg.split('-')[0] + '.txt'

    workspace_dir = getcwd()
    default_path = path.join(workspace_dir, data_dir)
    pkgpath = path.join(default_path, pkg)       # path to downloaded zipfile
    filepath = path.join(default_path, filename)  # path to extracted file

    if path.isfile(filepath):
        return path.join(data_dir, filename)
    try:
        response = requests.get(url, headers={'User-Agent': 'openfl'})
        if response.status_code == 200:
            with open(pkgpath, 'wb') as f:
                f.write(response.content)
        else:
            print(f'Error while downloading {pkg} from {url}: Aborting!')
            exit()
    except Exception:
        print(f'Error while downloading {pkg} from {url}: Aborting!')
        exit()

    try:
        with ZipFile(pkgpath, 'r') as z:
            z.extract(filename, default_path)
    except Exception:
        print(f'Error while extracting {pkgpath}: Aborting!')
        exit()

    if path.isfile(filepath):
        remove(pkgpath)
        return path.join(data_dir, filename)
    else:
        return ''


def import_raw_data_(
    data_path: str = '',
    num_samples: int = 0
) -> Tuple[Dict[str, int], np.ndarray, np.ndarray, np.ndarray]:
    """Import data.

    Returns:
      dict: variable details
      numpy.ndarray: encoder input data
      numpy.ndarray: decoder input data
      numpy.ndarray: decoder labels
    """
    # Vectorize the data.
    input_texts = []
    target_texts = []
    input_characters = set()
    target_characters = set()
    with open(data_path, 'r', encoding='utf-8') as f:
        lines = f.read().split('\n')
    for line in lines[: min(num_samples, len(lines) - 1)]:
        input_text, target_text, _ = line.split('\t')
        # We use 'tab' as the 'start sequence' character
        # for the targets, and '\n' as 'end sequence' character.
        target_text = '\t' + target_text + '\n'
        input_texts.append(input_text)
        target_texts.append(target_text)
        for char in input_text:
            if char not in input_characters:
                input_characters.add(char)
        for char in target_text:
            if char not in target_characters:
                target_characters.add(char)

    input_characters = sorted(input_characters)
    target_characters = sorted(target_characters)
    num_encoder_tokens = len(input_characters)
    num_decoder_tokens = len(target_characters)
    max_encoder_seq_length = max([len(txt) for txt in input_texts])
    max_decoder_seq_length = max([len(txt) for txt in target_texts])

    details = {'num_samples': len(input_texts),
               'num_encoder_tokens': num_encoder_tokens,
               'num_decoder_tokens': num_decoder_tokens,
               'max_encoder_seq_length': max_encoder_seq_length,
               'max_decoder_seq_length': max_decoder_seq_length}

    input_token_index = {char: i for i, char in enumerate(input_characters)}
    target_token_index = {char: i for i, char in enumerate(target_characters)}

    encoder_input_data = np.zeros(
        (len(input_texts), max_encoder_seq_length, num_encoder_tokens), dtype='float32')

    decoder_input_data = np.zeros(
        (len(input_texts), max_decoder_seq_length, num_decoder_tokens), dtype='float32')

    decoder_target_data = np.zeros(
        (len(input_texts), max_decoder_seq_length, num_decoder_tokens), dtype='float32')

    for i, (input_text, target_text) in enumerate(zip(input_texts, target_texts)):
        for t, char in enumerate(input_text):
            encoder_input_data[i, t, input_token_index[char]] = 1.0
        encoder_input_data[i, t + 1:, input_token_index[' ']] = 1.0
        for t, char in enumerate(target_text):
            # decoder_target_data is ahead of decoder_input_data by one timestep
            decoder_input_data[i, t, target_token_index[char]] = 1.0
            if t > 0:
                # decoder_target_data will be ahead by one timestep
                # and will not include the start character.
                decoder_target_data[i, t - 1, target_token_index[char]] = 1.0
        decoder_input_data[i, t + 1:, target_token_index[' ']] = 1.0
        decoder_target_data[i, t:, target_token_index[' ']] = 1.0

    logger.info(f'[DL]-import_raw_data: Number of samples = {len(input_texts)}')
    logger.info(f'[DL]-import_raw_data: Number of unique input tokens = {num_encoder_tokens}')
    logger.info(f'[DL]-import_raw_data: '
                f'Number of unique decoder tokens = {num_decoder_tokens}')

    logger.info(f'[DL]-import_raw_data: '
                f'Max sequence length for inputs = {max_encoder_seq_length}')

    logger.info(f'[DL]-import_raw_data: '
                f'Max sequence length for outputs = {max_decoder_seq_length}')

    logger.info(f'[DL]-import_raw_data: encoder_input_data = {encoder_input_data.shape}')
    logger.info(f'[DL]-import_raw_data: decoder_input_data = {decoder_input_data.shape}')
    logger.info(f'[DL]-import_raw_data: decoder_target_data = {decoder_target_data.shape}')

    return details, encoder_input_data, decoder_input_data, decoder_target_data


def get_datasets_(encoder_input_data: np.ndarray, decoder_input_data: np.ndarray,
                  decoder_target_data: np.ndarray,
                  num_samples: int, split_ratio: float) -> Dict[str, np.ndarray]:
    """Create train/val.

    Returns:
      dict: Results, containing the train-valid split of the dataset (split_ratio = 0.2)
    """
    import random

    random.seed(42)
    train_indexes = random.sample(range(num_samples), int(num_samples * (1 - split_ratio)))
    valid_indexes = np.delete(range(num_samples), train_indexes)

    # Dataset creation (2 inputs <encoder,decoder>, 1 output <decoder_target>)
    encoder_train_input = encoder_input_data[train_indexes, :, :]
    decoder_train_input = decoder_input_data[train_indexes, :, :]
    decoder_train_labels = decoder_target_data[train_indexes, :, :]

    encoder_valid_input = encoder_input_data[valid_indexes, :, :]
    decoder_valid_input = decoder_input_data[valid_indexes, :, :]
    decoder_valid_labels = decoder_target_data[valid_indexes, :, :]

    results = {'encoder_train_input': encoder_train_input,
               'decoder_train_input': decoder_train_input,
               'decoder_train_labels': decoder_train_labels,
               'encoder_valid_input': encoder_valid_input,
               'decoder_valid_input': decoder_valid_input,
               'decoder_valid_labels': decoder_valid_labels}

    logger.info(f'[DL]get_datasets: encoder_train_input = {encoder_train_input.shape}')
    logger.info(f'[DL]get_datasets: decoder_train_labels= {decoder_train_labels.shape}')

    return results


def load_shard(
    collaborator_count: int, shard_num: str, data_path: str,
    num_samples: int, split_ratio: float
) -> Tuple[Tuple[np.ndarray, ...], Tuple[np.ndarray, ...], Dict[str, int]]:
    """Load data-shards.

    Returns:
      Tuple: ( numpy.ndarray: X_train_encoder,
               numpy.ndarray: X_train_decoder,
               numpy.ndarray: y_train)
      Tuple: ( numpy.ndarray: X_valid_encoder,
               numpy.ndarray: X_valid_decoder,
               numpy.ndarray: y_valid)
      Dict: details, from DataLoader_utils.get_datasets
    """
    details, encoder_input_data, decoder_input_data, decoder_target_data = import_raw_data_(
        data_path,
        num_samples
    )

    train_val_dataset = get_datasets_(encoder_input_data, decoder_input_data,
                                      decoder_target_data, num_samples, split_ratio)
    # Get the data shards
    shard_num = int(shard_num)
    X_train_encoder = train_val_dataset['encoder_train_input'][shard_num::collaborator_count]
    X_train_decoder = train_val_dataset['decoder_train_input'][shard_num::collaborator_count]
    y_train = train_val_dataset['decoder_train_labels'][shard_num::collaborator_count]

    X_valid_encoder = train_val_dataset['encoder_valid_input'][shard_num::collaborator_count]
    X_valid_decoder = train_val_dataset['decoder_valid_input'][shard_num::collaborator_count]
    y_valid = train_val_dataset['decoder_valid_labels'][shard_num::collaborator_count]

    logger.info(f'[DL]load_shard: X_train_encoder = {X_train_encoder.shape}')
    logger.info(f'[DL]load_shard: y_train = {y_train.shape}')

    return (
        (X_train_encoder, X_train_decoder, y_train),
        (X_valid_encoder, X_valid_decoder, y_valid),
        details
    )



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Noisy-Sin Shard Descriptor."""

from typing import List

import numpy as np

from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor


class LinRegSD(ShardDescriptor):
    """Shard descriptor class."""

    def __init__(self, rank: int, n_samples: int = 10, noise: float = 0.15) -> None:
        """
        Initialize LinReg Shard Descriptor.

        This Shard Descriptor generate random data. Sample features are
        floats between pi/3 and 5*pi/3, and targets are calculated
        calculated as sin(feature) + normal_noise.
        """
        np.random.seed(rank)  # Setting seed for reproducibility
        self.n_samples = max(n_samples, 5)
        self.interval = 240
        self.x_start = 60
        x = np.random.rand(n_samples, 1) * self.interval + self.x_start
        x *= np.pi / 180
        y = np.sin(x) + np.random.normal(0, noise, size=(n_samples, 1))
        self.data = np.concatenate((x, y), axis=1)

    def get_dataset(self, dataset_type: str) -> np.ndarray:
        """
        Return a shard dataset by type.

        A simple list with elements (x, y) implemets the Shard Dataset interface.
        """
        if dataset_type == 'train':
            return self.data[:self.n_samples // 2]
        elif dataset_type == 'val':
            return self.data[self.n_samples // 2:]
        else:
            pass

    @property
    def sample_shape(self) -> List[str]:
        """Return the sample shape info."""
        (*x, _) = self.data[0]
        return [str(i) for i in np.array(x, ndmin=1).shape]

    @property
    def target_shape(self) -> List[str]:
        """Return the target shape info."""
        (*_, y) = self.data[0]
        return [str(i) for i in np.array(y, ndmin=1).shape]

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return 'Allowed dataset types are `train` and `val`'



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Assigner package."""

from .assigner import Assigner
from .random_grouped_assigner import RandomGroupedAssigner
from .static_grouped_assigner import StaticGroupedAssigner
from .adaboost_assigner import AdaBoostAssigner


__all__ = [
    'Assigner',
    'RandomGroupedAssigner',
    'StaticGroupedAssigner',
    'AdaBoostAssigner'
]



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Image transform tools."""

import math
import random

from PIL import Image


class ResizeRandomCropping:
    """
    With a probability, first increase image size to (1 + 1/8), and then perform random crop.

    Args:
        height (int): target height.
        width (int): target width.
        p (float): probability of performing this transformation. Default: 0.5.
    """

    def __init__(self, height, width, p=0.5, interpolation=Image.BILINEAR):
        """Initialize cropping."""
        self.height = height
        self.width = width
        self.p = p
        self.interpolation = interpolation

    def __call__(self, img):
        """
        Call of cropping.

        Args:
            img (PIL Image): Image to be cropped.
        Returns:
            PIL Image: Cropped image.
        """
        if random.uniform(0, 1) >= self.p:
            return img.resize((self.width, self.height), self.interpolation)

        new_width, new_height = int(round(self.width * 1.125)), int(round(self.height * 1.125))
        resized_img = img.resize((new_width, new_height), self.interpolation)
        x_maxrange = new_width - self.width
        y_maxrange = new_height - self.height
        x1 = int(round(random.uniform(0, x_maxrange)))
        y1 = int(round(random.uniform(0, y_maxrange)))
        cropped_img = resized_img.crop((x1, y1, x1 + self.width, y1 + self.height))

        return cropped_img


class RandomErasing:
    """
    Randomly selects a rectangle region in an image and erases its pixels.

    'Random Erasing Data Augmentation' by Zhong et al.
    See https://arxiv.org/pdf/1708.04896.pdf

    Args:
         probability: The probability that the Random Erasing operation will be performed.
         sl: Minimum proportion of erased area against input image.
         sh: Maximum proportion of erased area against input image.
         r1: Minimum aspect ratio of erased area.
         mean: Erasing value.
    """

    def __init__(self, probability=0.5, sl=0.02, sh=0.4, r1=0.3, mean=None):
        """Initialize Erasing."""
        if not mean:
            mean = [0.4914, 0.4822, 0.4465]

        self.probability = probability
        self.mean = mean
        self.sl = sl
        self.sh = sh
        self.r1 = r1

    def __call__(self, img):
        """Call of Erasing."""
        if random.uniform(0, 1) >= self.probability:
            return img

        for _attempt in range(100):
            area = img.size()[1] * img.size()[2]

            target_area = random.uniform(self.sl, self.sh) * area
            aspect_ratio = random.uniform(self.r1, 1 / self.r1)

            h = int(round(math.sqrt(target_area * aspect_ratio)))
            w = int(round(math.sqrt(target_area / aspect_ratio)))

            if w < img.size()[2] and h < img.size()[1]:
                x1 = random.randint(0, img.size()[1] - h)
                y1 = random.randint(0, img.size()[2] - w)
                if img.size()[0] == 3:
                    img[0, x1:x1 + h, y1:y1 + w] = self.mean[0]
                    img[1, x1:x1 + h, y1:y1 + w] = self.mean[1]
                    img[2, x1:x1 + h, y1:y1 + w] = self.mean[2]
                else:
                    img[0, x1:x1 + h, y1:y1 + w] = self.mean[0]
                return img

        return img



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Adagrad optimizer module."""

from typing import Dict
from typing import Optional

import numpy as np

from .base_optimizer import Optimizer


class NumPyAdagrad(Optimizer):
    """Adagrad optimizer implementation.

    Original paper: http://jmlr.org/papers/v12/duchi11a.html
    """

    def __init__(
        self,
        *,
        params: Optional[Dict[str, np.ndarray]] = None,
        model_interface=None,
        learning_rate: float = 0.01,
        initial_accumulator_value: float = 0.1,
        epsilon: float = 1e-10,
    ) -> None:
        """Initialize.

        Args:
            params: Parameters to be stored for optimization.
            model_interface: Model interface instance to provide parameters.
            learning_rate: Tuning parameter that determines
                the step size at each iteration.
            initial_accumulator_value: Initial value for squared gradients.
            epsilon: Value for computational stability.
        """
        super().__init__()

        if model_interface is None and params is None:
            raise ValueError('Should provide one of the params or model_interface')

        if learning_rate < 0:
            raise ValueError(
                f'Invalid learning rate: {learning_rate}. Learning rate must be >= 0.')
        if initial_accumulator_value < 0:
            raise ValueError(
                f'Invalid initial_accumulator_value value: {initial_accumulator_value}.'
                'Initial accumulator value must be >= 0.')
        if epsilon <= 0:
            raise ValueError(
                f'Invalid epsilon value: {epsilon}. Epsilon avalue must be > 0.')

        self.params = params

        if params is None and model_interface is not None:
            self._set_params_from_model(model_interface)

        self.learning_rate = learning_rate
        self.initial_accumulator_value = initial_accumulator_value
        self.epsilon = epsilon

        self.grads_squared = {}
        for param_name in self.params:
            self.grads_squared[param_name] = np.full_like(self.params[param_name],
                                                          self.initial_accumulator_value)

    def _update_param(self, grad_name: str, grad: np.ndarray) -> None:
        """Update papams by given gradients."""
        self.params[grad_name] -= (self.learning_rate * grad
                                   / (np.sqrt(self.grads_squared[grad_name]) + self.epsilon))

    def step(self, gradients: Dict[str, np.ndarray]) -> None:
        """
        Perform a single step for parameter update.

        Implement Adagrad optimizer weights update rule.

        Args:
            gradients: Partial derivatives with respect to optimized parameters.
        """
        for grad_name in gradients:
            if grad_name not in self.grads_squared:
                raise KeyError(f"Key {grad_name} doesn't exist in optimized parameters")

            grad = gradients[grad_name]
            self.grads_squared[grad_name] = self.grads_squared[grad_name] + grad**2
            self._update_param(grad_name, grad)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""
Base classes for developing a ke.Model() Federated Learning model.

You may copy this file as the starting point of your own keras model.
"""
from warnings import catch_warnings
from warnings import simplefilter

import numpy as np

from openfl.utilities import change_tags
from openfl.utilities import Metric
from openfl.utilities import split_tensor_dict_for_holdouts
from openfl.utilities import TensorKey
from .runner import TaskRunner

with catch_warnings():
    simplefilter(action='ignore')
    import tensorflow as tf
    import tensorflow.keras as ke


class KerasTaskRunner(TaskRunner):
    """The base model for Keras models in the federation."""

    def __init__(self, **kwargs):
        """
        Initialize.

        Args:
            **kwargs: Additional parameters to pass to the function
        """
        super().__init__(**kwargs)

        self.model = ke.Model()

        self.model_tensor_names = []

        # this is a map of all of the required tensors for each of the public
        # functions in KerasTaskRunner
        self.required_tensorkeys_for_function = {}
        ke.backend.clear_session()

    def rebuild_model(self, round_num, input_tensor_dict, validation=False):
        """
        Parse tensor names and update weights of model. Handles the optimizer treatment.

        Returns
        -------
        None
        """
        if self.opt_treatment == 'RESET':
            self.reset_opt_vars()
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=False)
        elif (round_num > 0 and self.opt_treatment == 'CONTINUE_GLOBAL'
              and not validation):
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=True)
        else:
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=False)

    def train(self, col_name, round_num, input_tensor_dict,
              metrics, epochs=1, batch_size=1, **kwargs):
        """
        Perform the training.

        Is expected to perform draws randomly, without replacement until data is exausted.
        Then data is replaced and shuffled and draws continue.

        Returns
        -------
        dict
            'TensorKey: nparray'
        """
        if metrics is None:
            raise KeyError('metrics must be defined')

        # rebuild model with updated weights
        self.rebuild_model(round_num, input_tensor_dict)
        for epoch in range(epochs):
            self.logger.info(f'Run {epoch} epoch of {round_num} round')
            results = self.train_iteration(self.data_loader.get_train_loader(batch_size),
                                           metrics=metrics,
                                           **kwargs)

        # output metric tensors (scalar)
        origin = col_name
        tags = ('trained',)
        output_metric_dict = {
            TensorKey(
                metric_name, origin, round_num, True, ('metric',)
            ): metric_value
            for (metric_name, metric_value) in results
        }

        # output model tensors (Doesn't include TensorKey)
        output_model_dict = self.get_tensor_dict(with_opt_vars=True)
        global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
            self.logger, output_model_dict,
            **self.tensor_dict_split_fn_kwargs
        )

        # create global tensorkeys
        global_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num, False, tags):
                nparray for tensor_name, nparray in global_model_dict.items()
        }
        # create tensorkeys that should stay local
        local_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num, False, tags):
                nparray for tensor_name, nparray in local_model_dict.items()
        }
        # the train/validate aggregated function of the next round will look
        # for the updated model parameters.
        # this ensures they will be resolved locally
        next_local_tensorkey_model_dict = {
            TensorKey(
                tensor_name, origin, round_num + 1, False, ('model',)
            ): nparray for tensor_name, nparray in local_model_dict.items()
        }

        global_tensor_dict = {
            **output_metric_dict,
            **global_tensorkey_model_dict
        }
        local_tensor_dict = {
            **local_tensorkey_model_dict,
            **next_local_tensorkey_model_dict
        }

        # update the required tensors if they need to be pulled from the
        # aggregator
        # TODO this logic can break if different collaborators have different
        # roles between rounds.
        # for example, if a collaborator only performs validation in the first
        # round but training in the second, it has no way of knowing the
        # optimizer state tensor names to request from the aggregator because
        # these are only created after training occurs. A work around could
        # involve doing a single epoch of training on random data to get the
        # optimizer names, and then throwing away the model.
        if self.opt_treatment == 'CONTINUE_GLOBAL':
            self.initialize_tensorkeys_for_functions(with_opt_vars=True)

        return global_tensor_dict, local_tensor_dict

    def train_iteration(self, batch_generator, metrics: list = None, **kwargs):
        """Train single epoch.

        Override this function for custom training.

        Args:
            batch_generator: Generator of training batches.
                Each batch is a tuple of N train images and N train labels
                where N is the batch size of the DataLoader of the current TaskRunner instance.

            epochs: Number of epochs to train.
            metrics: Names of metrics to save.
        """
        if metrics is None:
            metrics = []
        # TODO Currently assuming that all metrics are defined at
        #  initialization (build_model).
        #  If metrics are added (i.e. not a subset of what was originally
        #  defined) then the model must be recompiled.
        model_metrics_names = self.model.metrics_names

        # TODO if there are new metrics in the flplan that were not included
        #  in the originally
        #  compiled model, that behavior is not currently handled.
        for param in metrics:
            if param not in model_metrics_names:
                raise ValueError(
                    f'KerasTaskRunner does not support specifying new metrics. '
                    f'Param_metrics = {metrics}, model_metrics_names = {model_metrics_names}'
                )

        history = self.model.fit(batch_generator,
                                 verbose=1,
                                 **kwargs)
        results = []
        for metric in metrics:
            value = np.mean([history.history[metric]])
            results.append(Metric(name=metric, value=np.array(value)))
        return results

    def validate(self, col_name, round_num, input_tensor_dict, **kwargs):
        """
        Run the trained model on validation data; report results.

        Parameters
        ----------
        input_tensor_dict : either the last aggregated or locally trained model

        Returns
        -------
        output_tensor_dict : {TensorKey: nparray} (these correspond to acc,
         precision, f1_score, etc.)
        """
        if 'batch_size' in kwargs:
            batch_size = kwargs['batch_size']
        else:
            batch_size = 1

        self.rebuild_model(round_num, input_tensor_dict, validation=True)
        param_metrics = kwargs['metrics']

        vals = self.model.evaluate(
            self.data_loader.get_valid_loader(batch_size),
            verbose=1
        )
        model_metrics_names = self.model.metrics_names
        if type(vals) is not list:
            vals = [vals]
        ret_dict = dict(zip(model_metrics_names, vals))

        # TODO if there are new metrics in the flplan that were not included in
        #  the originally compiled model, that behavior is not currently
        #  handled.
        for param in param_metrics:
            if param not in model_metrics_names:
                raise ValueError(
                    f'KerasTaskRunner does not support specifying new metrics. '
                    f'Param_metrics = {param_metrics}, model_metrics_names = {model_metrics_names}'
                )

        origin = col_name
        suffix = 'validate'
        if kwargs['apply'] == 'local':
            suffix += '_local'
        else:
            suffix += '_agg'
        tags = ('metric',)
        tags = change_tags(tags, add_field=suffix)
        output_tensor_dict = {
            TensorKey(metric, origin, round_num, True, tags):
                np.array(ret_dict[metric])
            for metric in param_metrics}

        return output_tensor_dict, {}

    def save_native(self, filepath):
        """Save model."""
        self.model.save(filepath)

    def load_native(self, filepath):
        """Load model."""
        self.model = ke.models.load_model(filepath)

    @staticmethod
    def _get_weights_names(obj):
        """
        Get the list of weight names.

        Parameters
        ----------
        obj : Model or Optimizer
            The target object that we want to get the weights.

        Returns
        -------
        dict
            The weight name list
        """
        weight_names = [weight.name for weight in obj.weights]
        return weight_names

    @staticmethod
    def _get_weights_dict(obj, suffix=''):
        """
        Get the dictionary of weights.

        Parameters
        ----------
        obj : Model or Optimizer
            The target object that we want to get the weights.

        Returns
        -------
        dict
            The weight dictionary.
        """
        weights_dict = {}
        weight_names = [weight.name for weight in obj.weights]
        weight_values = obj.get_weights()
        for name, value in zip(weight_names, weight_values):
            weights_dict[name + suffix] = value
        return weights_dict

    @staticmethod
    def _set_weights_dict(obj, weights_dict):
        """Set the object weights with a dictionary.

        The obj can be a model or an optimizer.

        Args:
            obj (Model or Optimizer): The target object that we want to set
            the weights.
            weights_dict (dict): The weight dictionary.

        Returns:
            None
        """
        weight_names = [weight.name for weight in obj.weights]
        weight_values = [weights_dict[name] for name in weight_names]
        obj.set_weights(weight_values)

    def get_tensor_dict(self, with_opt_vars, suffix=''):
        """
        Get the model weights as a tensor dictionary.

        Parameters
        ----------
        with_opt_vars : bool
            If we should include the optimizer's status.
        suffix : string
            Universally

        Returns:
            dict: The tensor dictionary.
        """
        model_weights = self._get_weights_dict(self.model, suffix)

        if with_opt_vars:
            opt_weights = self._get_weights_dict(self.model.optimizer, suffix)

            model_weights.update(opt_weights)
            if len(opt_weights) == 0:
                self.logger.debug(
                    "WARNING: We didn't find variables for the optimizer.")
        return model_weights

    def set_tensor_dict(self, tensor_dict, with_opt_vars):
        """
        Set the model weights with a tensor dictionary.

        Args:
            tensor_dict: the tensor dictionary
            with_opt_vars (bool): True = include the optimizer's status.
        """
        if with_opt_vars is False:
            # It is possible to pass in opt variables from the input tensor dict
            # This will make sure that the correct layers are updated
            model_weight_names = [weight.name for weight in self.model.weights]
            model_weights_dict = {
                name: tensor_dict[name] for name in model_weight_names
            }
            self._set_weights_dict(self.model, model_weights_dict)
        else:
            model_weight_names = [
                weight.name for weight in self.model.weights
            ]
            model_weights_dict = {
                name: tensor_dict[name] for name in model_weight_names
            }
            opt_weight_names = [
                weight.name for weight in self.model.optimizer.weights
            ]
            opt_weights_dict = {
                name: tensor_dict[name] for name in opt_weight_names
            }
            self._set_weights_dict(self.model, model_weights_dict)
            self._set_weights_dict(self.model.optimizer, opt_weights_dict)

    def reset_opt_vars(self):
        """
        Reset optimizer variables.

        Resets the optimizer variables

        """
        for var in self.model.optimizer.variables():
            var.assign(tf.zeros_like(var))
        self.logger.debug('Optimizer variables reset')

    def set_required_tensorkeys_for_function(self, func_name,
                                             tensor_key, **kwargs):
        """
        Set the required tensors for specified function that could be called as part of a task.

        By default, this is just all of the layers and optimizer of the model.
         Custom tensors should be added to this function

        Parameters
        ----------
        func_name: string
        tensor_key: TensorKey (namedtuple)
        **kwargs: Any function arguments {}

        Returns
        -------
        None
        """
        # TODO there should be a way to programmatically iterate through all
        #  of the methods in the class and declare the tensors.
        # For now this is done manually

        if func_name == 'validate':
            # Should produce 'apply=global' or 'apply=local'
            local_model = 'apply' + kwargs['apply']
            self.required_tensorkeys_for_function[func_name][
                local_model].append(tensor_key)
        else:
            self.required_tensorkeys_for_function[func_name].append(tensor_key)

    def get_required_tensorkeys_for_function(self, func_name, **kwargs):
        """
        Get the required tensors for specified function that could be called as part of a task.

        By default, this is just all of the layers and optimizer of the model.

        Parameters
        ----------
        None

        Returns
        -------
        List
            [TensorKey]
        """
        if func_name == 'validate':
            local_model = 'apply=' + str(kwargs['apply'])
            return self.required_tensorkeys_for_function[func_name][local_model]
        else:
            return self.required_tensorkeys_for_function[func_name]

    def update_tensorkeys_for_functions(self):
        """
        Update the required tensors for all publicly accessible methods \
            that could be called as part of a task.

        By default, this is just all of the layers and optimizer of the model.
        Custom tensors should be added to this function

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        # TODO complete this function. It is only needed for opt_treatment,
        #  and making the model stateless

        # Minimal required tensors for train function
        model_layer_names = self._get_weights_names(self.model)
        opt_names = self._get_weights_names(self.model.optimizer)
        tensor_names = model_layer_names + opt_names
        self.logger.debug(f'Updating model tensor names: {tensor_names}')
        self.required_tensorkeys_for_function['train'] = [
            TensorKey(tensor_name, 'GLOBAL', 0, ('model',))
            for tensor_name in tensor_names
        ]

        # Validation may be performed on local or aggregated (global) model,
        # so there is an extra lookup dimension for kwargs
        self.required_tensorkeys_for_function['validate'] = {}
        self.required_tensorkeys_for_function['validate']['local_model=True'] = [
            TensorKey(tensor_name, 'LOCAL', 0, ('trained',))
            for tensor_name in tensor_names
        ]
        self.required_tensorkeys_for_function['validate']['local_model=False'] = [
            TensorKey(tensor_name, 'GLOBAL', 0, ('model',))
            for tensor_name in tensor_names
        ]

    def initialize_tensorkeys_for_functions(self, with_opt_vars=False):
        """
        Set the required tensors for all publicly accessible methods \
            that could be called as part of a task.

        By default, this is just all of the layers and optimizer of the model.
        Custom tensors should be added to this function

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        # TODO there should be a way to programmatically iterate through all
        #  of the methods in the class and declare the tensors.
        # For now this is done manually

        output_model_dict = self.get_tensor_dict(with_opt_vars=with_opt_vars)
        global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
            self.logger, output_model_dict,
            **self.tensor_dict_split_fn_kwargs
        )
        if not with_opt_vars:
            global_model_dict_val = global_model_dict
            local_model_dict_val = local_model_dict
        else:
            output_model_dict = self.get_tensor_dict(with_opt_vars=False)
            global_model_dict_val, local_model_dict_val = split_tensor_dict_for_holdouts(
                self.logger,
                output_model_dict,
                **self.tensor_dict_split_fn_kwargs
            )

        self.required_tensorkeys_for_function['train'] = [
            TensorKey(tensor_name, 'GLOBAL', 0, False, ('model',))
            for tensor_name in global_model_dict
        ]
        self.required_tensorkeys_for_function['train'] += [
            TensorKey(tensor_name, 'LOCAL', 0, False, ('model',))
            for tensor_name in local_model_dict
        ]

        # Validation may be performed on local or aggregated (global) model,
        # so there is an extra lookup dimension for kwargs
        self.required_tensorkeys_for_function['validate'] = {}
        # TODO This is not stateless. The optimizer will not be
        self.required_tensorkeys_for_function['validate']['apply=local'] = [
            TensorKey(tensor_name, 'LOCAL', 0, False, ('trained',))
            for tensor_name in {
                **global_model_dict_val,
                **local_model_dict_val
            }
        ]
        self.required_tensorkeys_for_function['validate']['apply=global'] = [
            TensorKey(tensor_name, 'GLOBAL', 0, False, ('model',))
            for tensor_name in global_model_dict_val
        ]
        self.required_tensorkeys_for_function['validate']['apply=global'] += [
            TensorKey(tensor_name, 'LOCAL', 0, False, ('model',))
            for tensor_name in local_model_dict_val
        ]



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Utilities module."""

import hashlib
import ipaddress
import logging
import os
import re
from collections.abc import Callable
from functools import partial
from socket import getfqdn
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
from dynaconf import Dynaconf
from tqdm import tqdm


def getfqdn_env(name: str = '') -> str:
    """
    Get the system FQDN, with priority given to environment variables.

    Args:
        name: The name from which to extract the FQDN.

    Returns:
        The FQDN of the system.
    """
    fqdn = os.environ.get('FQDN', None)
    if fqdn is not None:
        return fqdn
    return getfqdn(name)


def is_fqdn(hostname: str) -> bool:
    """https://en.m.wikipedia.org/wiki/Fully_qualified_domain_name."""
    if not 1 < len(hostname) < 253:
        return False

    # Remove trailing dot
    hostname.rstrip('.')

    #  Split hostname into list of DNS labels
    labels = hostname.split('.')

    #  Define pattern of DNS label
    #  Can begin and end with a number or letter only
    #  Can contain hyphens, a-z, A-Z, 0-9
    #  1 - 63 chars allowed
    fqdn = re.compile(r'^[a-z0-9]([a-z-0-9-]{0,61}[a-z0-9])?$', re.IGNORECASE)  # noqa FS003

    # Check that all labels match that pattern.
    return all(fqdn.match(label) for label in labels)


def is_api_adress(address: str) -> bool:
    """Validate ip address value."""
    try:
        ipaddress.ip_address(address)
        return True
    except ValueError:
        return False


def add_log_level(level_name, level_num, method_name=None):
    """
    Add a new logging level to the logging module.

    Args:
        level_name: name of log level.
        level_num: log level value.
        method_name: log method wich will use new log level (default = level_name.lower())

    """
    if not method_name:
        method_name = level_name.lower()

    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)


def split_tensor_dict_into_floats_and_non_floats(tensor_dict):
    """
    Split the tensor dictionary into float and non-floating point values.

    Splits a tensor dictionary into float and non-float values.

    Args:
        tensor_dict: A dictionary of tensors

    Returns:
        Two dictionaries: the first contains all of the floating point tensors
        and the second contains all of the non-floating point tensors

    """
    float_dict = {}
    non_float_dict = {}
    for k, v in tensor_dict.items():
        if np.issubdtype(v.dtype, np.floating):
            float_dict[k] = v
        else:
            non_float_dict[k] = v
    return float_dict, non_float_dict


def split_tensor_dict_by_types(tensor_dict, keep_types):
    """
    Split the tensor dictionary into supported and not supported types.

    Args:
        tensor_dict: A dictionary of tensors
        keep_types: An iterable of supported types
    Returns:
        Two dictionaries: the first contains all of the supported tensors
        and the second contains all of the not supported tensors

    """
    keep_dict = {}
    holdout_dict = {}
    for k, v in tensor_dict.items():
        if any([np.issubdtype(v.dtype, type_) for type_ in keep_types]):
            keep_dict[k] = v
        else:
            holdout_dict[k] = v
    return keep_dict, holdout_dict


def split_tensor_dict_for_holdouts(logger, tensor_dict,
                                   keep_types=(np.floating, np.integer),
                                   holdout_tensor_names=()):
    """
    Split a tensor according to tensor types.

    Args:
        logger: The log object
        tensor_dict: A dictionary of tensors
        keep_types: A list of types to keep in dictionary of tensors
        holdout_tensor_names: A list of tensor names to extract from the
         dictionary of tensors

    Returns:
        Two dictionaries: the first is the original tensor dictionary minus
        the holdout tenors and the second is a tensor dictionary with only the
        holdout tensors

    """
    # initialization
    tensors_to_send = tensor_dict.copy()
    holdout_tensors = {}

    # filter by-name tensors from tensors_to_send and add to holdout_tensors
    # (for ones not already held out becuase of their type)
    for tensor_name in holdout_tensor_names:
        if tensor_name not in holdout_tensors.keys():
            try:
                holdout_tensors[tensor_name] = tensors_to_send.pop(tensor_name)
            except KeyError:
                logger.warn(f'tried to remove tensor: {tensor_name} not present '
                            f'in the tensor dict')
                continue

    # filter holdout_types from tensors_to_send and add to holdout_tensors
    tensors_to_send, not_supported_tensors_dict = split_tensor_dict_by_types(
        tensors_to_send,
        keep_types
    )
    holdout_tensors = {**holdout_tensors, **not_supported_tensors_dict}

    return tensors_to_send, holdout_tensors


def validate_file_hash(file_path, expected_hash, chunk_size=8192):
    """Validate SHA384 hash for file specified.

    Args:
        file_path(path-like): path-like object giving the pathname
            (absolute or relative to the current working directory)
            of the file to be opened or an integer file descriptor of the file to be wrapped.
        expected_hash(str): hash string to compare with.
        hasher(_Hash): hash algorithm. Default value: `hashlib.sha384()`
        chunk_size(int): Buffer size for file reading.
    """
    h = hashlib.sha384()
    with open(file_path, 'rb') as file:
        # Reading is buffered, so we can read smaller chunks.
        while True:
            chunk = file.read(chunk_size)
            if not chunk:
                break
            h.update(chunk)

    if h.hexdigest() != expected_hash:
        raise SystemError('ZIP File hash doesn\'t match expected file hash.')


def tqdm_report_hook():
    """Visualize downloading."""

    def report_hook(pbar, count, block_size, total_size):
        """Update progressbar."""
        if pbar.total is None and total_size:
            pbar.total = total_size
        progress_bytes = count * block_size
        pbar.update(progress_bytes - pbar.n)

    pbar = tqdm(total=None)
    return partial(report_hook, pbar)


def merge_configs(
        overwrite_dict: Optional[dict] = None,
        value_transform: Optional[List[Tuple[str, Callable]]] = None,
        **kwargs,
) -> Dynaconf:
    """Create Dynaconf settings, merge its with `overwrite_dict` and validate result."""
    settings = Dynaconf(**kwargs, YAML_LOADER='safe_load')
    if overwrite_dict:
        for key, value in overwrite_dict.items():
            if value is not None or settings.get(key) is None:
                settings.set(key, value, merge=True)
    if value_transform:
        for key, operation in value_transform:
            value = settings.get(key)
            settings.set(key, operation(value))
    settings.validators.validate()
    return settings


def change_tags(tags, *, add_field=None, remove_field=None) -> Tuple[str, ...]:
    """Change tensor tags to add or remove fields.

    Args:
        tags(tuple): tensor tags.
        add_field(str): add a new tensor tag field.
        remove_field(str): remove a tensor tag field.
    """
    tags = list(set(tags))

    if add_field is not None and add_field not in tags:
        tags.append(add_field)
    if remove_field is not None:
        if remove_field in tags:
            tags.remove(remove_field)
        else:
            raise Exception(f'{remove_field} not in tags {tuple(tags)}')

    tags = tuple(sorted(tags))
    return tags



from openfl.plugins.frameworks_adapters.framework_adapter_interface import FrameworkAdapterPluginInterface


class GenericAdapter(FrameworkAdapterPluginInterface):
    """Generic adapter plugin class."""

    def __init__(self) -> None:
        """Initialize framework adapter."""
        pass

    @staticmethod
    def serialization_setup():
        """Prepare model for serialization (optional)."""
        pass

    @staticmethod
    def get_tensor_dict(model, optimizer=None) -> dict:
        """
        Extract tensor dict from a model and an optimizer.

        Returns:
        dict {weight name: numpy ndarray}
        """
        return {'generic_model': model}

    @staticmethod
    def set_tensor_dict(model, tensor_dict, optimizer=None, device='cpu'):
        """
        Set tensor dict from a model and an optimizer.

        Given a dict {weight name: numpy ndarray} sets weights to
        the model and optimizer objects inplace.
        """
        pass


# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Convenience Utilities for DataFrame."""

from types import MethodType

import numpy as np
import pandas as pd
from typing import Optional, Union

ROUND_PLACEHOLDER = 1000000

def _search(self, tensor_name: str = None, origin: str = None, 
            fl_round: int = None, metric: bool = None, tags: tuple = None
            ) -> pd.DataFrame:
    """
        Search the tensor_db dataframe based on:
            - tensor_name
            - origin
            - fl_round
            - metric
            -tags

        Returns a new dataframe that matched the query

        Args:
            tensor_name: The name of the tensor (or metric) to be searched
            origin:      Origin of the tensor
            fl_round:    Round the tensor is associated with
            metric:      Is the tensor a metric?
            tags:        Tuple of unstructured tags associated with the tensor

        Returns:
            pd.DataFrame : New dataframe that matches the search query from 
                           the tensor_db dataframe
    """
    df = None
    query_string = []
    if tensor_name is not None:
        query_string.append(f"(tensor_name == '{tensor_name}')")
    if origin is not None:
        query_string.append(f"(origin == '{origin}')")
    if fl_round is not None:
        query_string.append(f"(round == {fl_round})")
    if metric is not None:
        query_string.append(f"(report == {metric})")

    if len(query_string) > 0:
        query_string = (' and ').join(query_string)
        df = self.query(query_string)
    if tags is not None:
        if df is not None:
            df = df[df['tags'] == tags]
        else:
            df = self[self['tags'] == tags]

    if df is not None:
        return df
    else:
        return self


def _store(self, tensor_name: str = '_', origin: str = '_',
           fl_round: int = ROUND_PLACEHOLDER, metric: bool = False, 
           tags: tuple = ('_',), nparray: np.array = None, 
           overwrite: bool = True) -> None:
    """
        Convenience method to store a new tensor in the dataframe.

        Args:
            tensor_name [ optional ] : The name of the tensor (or metric) to be saved
            origin      [ optional ] : Origin of the tensor
            fl_round    [ optional ] : Round the tensor is associated with
            metric:     [ optional ] : Is the tensor a metric?
            tags:       [ optional ] : Tuple of unstructured tags associated with the tensor
            np.array    [ required ] : Value to store associated with the other included information (i.e. TensorKey info)
            overwrite:  [ optional ] : If the tensor is already present in the dataframe
                                       should it be overwritten?

        Returns:
            None
    """

    if nparray is None:
        print('nparray not provided. Nothing to store.')
        return
    idx = self[(self['tensor_name'] == tensor_name)
               & (self['origin'] == origin)
               & (self['round'] == str(fl_round))
               & (self['tags'] == tags)].index
    if len(idx) > 0:
        if not overwrite:
            return
        idx = idx[0]
    else:
        idx = self.shape[0]
    self.loc[idx] = \
            [tensor_name, origin, str(fl_round), metric, tags, nparray]


def _retrieve(self, tensor_name: str = '_', origin: str = '_',
              fl_round: int = ROUND_PLACEHOLDER, metric: bool = False, 
              tags: tuple = ('_',)) -> Optional[np.array]:
    """
        Convenience method to retrieve tensor from the dataframe.

        Args:
            tensor_name [ optional ] : The name of the tensor (or metric) to retrieve
            origin      [ optional ] : Origin of the tensor
            fl_round    [ optional ] : Round the tensor is associated with
            metric:     [ optional ] : Is the tensor a metric?
            tags:       [ optional ] : Tuple of unstructured tags associated with the tensor
                                       should it be overwritten?

        Returns:
            Optional[ np.array ]     : If there is a match, return the first row
    """

    df = self[(self['tensor_name'] == tensor_name)
                & (self['origin'] == origin)
                & (self['round'] == str(fl_round))
                & (self['report'] == metric)
                & (self['tags'] == tags)]['nparray']

    if len(df) > 0:
        return df.iloc[0]
    else:
        return None





# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Pipeline module."""

import numpy as np
from dill import dumps, loads

class Transformer:
    """Data transformation class."""

    def forward(self, data, **kwargs):
        """Forward pass data transformation.

        Implement the data transformation.

        Args:
            data:
            **kwargs: Additional parameters to pass to the function

        Returns:
            transformed_data
            metadata
        """
        raise NotImplementedError

    def backward(self, data, metadata, **kwargs):
        """Backward pass data transformation.

        Implement the data transformation needed when going the opposite
        direction to the forward method.

        Args:
            data:
            metadata:
            **kwargs: Additional parameters to pass to the function

        Returns:
            transformed_data
        """
        raise NotImplementedError


class Float32NumpyArrayToBytes(Transformer):
    """Converts float32 Numpy array to Bytes array."""

    def __init__(self, nn=True):
        """Initialize."""
        self.lossy = False
        self.nn = nn

    def forward(self, data, **kwargs):
        """Forward pass.

        Args:
            data:
            **kwargs: Additional arguments to pass to the function

        Returns:
            data_bytes:
            metadata:
        """
        # TODO: Warn when this casting is being performed.
        if self.nn or isinstance(data, np.ndarray):
            if data.dtype != np.float32:
                data = data.astype(np.float32)
            array_shape = data.shape
            # Better call it array_shape?
            metadata = {'int_list': list(array_shape)}
            data_bytes = data.tobytes(order='C')
        else:
            data_bytes = dumps(data)
            metadata = {'model': True}

        return data_bytes, metadata

    def backward(self, data, metadata, **kwargs):
        """Backward pass.

        Args:
            data:
            metadata:

        Returns:
            Numpy Array

        """
        if 'model' not in metadata or not metadata['model']:
            array_shape = tuple(metadata['int_list'])
            flat_array = np.frombuffer(data, dtype=np.float32)
            # For integer parameters we probably should unpack arrays
            # with shape (1,)
            result = np.reshape(flat_array, newshape=array_shape, order='C')
        else:
            result = loads(data)

        return result


class TransformationPipeline:
    """Data Transformer Pipeline Class.

    A sequential pipeline to transform (e.x. compress) data (e.x. layer of
    model_weights) as well as return metadata (if needed) for the
    reconstruction process carried out by the backward method.
    """

    # TODO: maybe put nn in **kwargs
    def __init__(self, transformers, nn=True, **kwargs):
        """Initialize.

        Args:
            transformers:
            **kwargs: Additional parameters to pass to the function
        """
        self.transformers = transformers
        self.nn = nn

    def forward(self, data, **kwargs):
        """Forward pass of pipeline data transformer.

        Args:
            data: Data to transform
            **kwargs: Additional parameters to pass to the function

        Returns:
            data:
            transformer_metadata:

        """
        transformer_metadata = []

        # dataformat::numpy::float.32
        # model proto:: a collection of tensor_dict proto
        # protobuff::-> a layer of weights
        # input::tensor_dict:{"layer1":np.array(float32, [128,3,28,28]),
        # "layer2": np.array()}
        # output::meta data::numpy::int array
        # (data, transformer_metadata)::(float32, dictionary o
        # key+float32 vlues)
        # input:: numpy_data (float32)
        # input:: (data(bytes), transformer_metadata_list::a list of dictionary
        # from int to float)

        if self.nn:
            data = data.copy()

        for transformer in self.transformers:
            data, metadata = transformer.forward(data=data, **kwargs)
            transformer_metadata.append(metadata)
        return data, transformer_metadata

    def backward(self, data, transformer_metadata, **kwargs):
        """Backward pass of pipeline data transformer.

        Args:
            data: Data to transform
            transformer_metadata:
            **kwargs: Additional parameters to pass to the function

        Returns:
            data:

        """
        for transformer in self.transformers[::-1]:
            data = transformer.backward(
                data=data, metadata=transformer_metadata.pop(), **kwargs)
        return data

    def is_lossy(self):
        """If any of the transformers are lossy, then the pipeline is lossy."""
        return any([transformer.lossy for transformer in self.transformers])



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Random grouped assigner module."""


import numpy as np

from .assigner import Assigner


class RandomGroupedAssigner(Assigner):
    r"""
    The task assigner maintains a list of tasks.

    Also it decides the policy for
    which collaborator should run those tasks
    There may be many types of policies implemented, but a natural place to
    start is with a:

    RandomGroupedAssigner  - Given a set of task groups, and a percentage,
                             assign that task group to that percentage
                             of collaborators in the federation. After
                             assigning the tasks to collaborator, those
                             tasks should be carried out each round (no
                             reassignment between rounds)
    GroupedAssigner -        Given task groups and a list of collaborators
                             that belong to that task group,
                             carry out tasks for each round of experiment

    Args:
        task_groups* (list of object): task groups to assign.

    Note:
        \* - Plan setting.
    """

    def __init__(self, task_groups, **kwargs):
        """Initialize."""
        self.task_groups = task_groups
        super().__init__(**kwargs)

    def define_task_assignments(self):
        """All of the logic to set up the map of tasks to collaborators is done here."""
        assert (np.abs(1.0 - np.sum([group['percentage']
                                     for group in self.task_groups])) < 0.01), (
            'Task group percentages must sum to 100%')

        # Start by finding all of the tasks in all specified groups
        self.all_tasks_in_groups = list({
            task
            for group in self.task_groups
            for task in group['tasks']
        })

        # Initialize the map of collaborators for a given task on a given round
        for task in self.all_tasks_in_groups:
            self.collaborators_for_task[task] = {
                i: [] for i in range(self.rounds)
            }

        for col in self.authorized_cols:
            self.collaborator_tasks[col] = {i: [] for i in range(self.rounds)}

        col_list_size = len(self.authorized_cols)
        for round_num in range(self.rounds):
            randomized_col_idx = np.random.choice(
                len(self.authorized_cols),
                len(self.authorized_cols),
                replace=False
            )
            col_idx = 0
            for group in self.task_groups:
                num_col_in_group = int(group['percentage'] * col_list_size)
                rand_col_group_list = [
                    self.authorized_cols[i] for i in
                    randomized_col_idx[col_idx:col_idx + num_col_in_group]
                ]
                self.task_group_collaborators[group['name']] = rand_col_group_list
                for col in rand_col_group_list:
                    self.collaborator_tasks[col][round_num] = group['tasks']
                # Now populate reverse lookup of tasks->group
                for task in group['tasks']:
                    # This should append the list of collaborators performing
                    # that task
                    self.collaborators_for_task[task][round_num] += rand_col_group_list
                col_idx += num_col_in_group
            assert (col_idx == col_list_size), 'Task groups were not divided properly'

    def get_tasks_for_collaborator(self, collaborator_name, round_number):
        """Get tasks for the collaborator specified."""
        return self.collaborator_tasks[collaborator_name][round_number]

    def get_collaborators_for_task(self, task_name, round_number):
        """Get collaborators for the task specified."""
        return self.collaborators_for_task[task_name][round_number]



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Envoy module."""

import logging
import time
import traceback
import uuid
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Optional
from typing import Type
from typing import Union

from click import echo

from openfl.federated import Plan
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor
from openfl.plugins.processing_units_monitor.cuda_device_monitor import CUDADeviceMonitor
from openfl.transport.grpc.director_client import ShardDirectorClient
from openfl.utilities.workspace import ExperimentWorkspace

logger = logging.getLogger(__name__)

DEFAULT_RETRY_TIMEOUT_IN_SECONDS = 5


class Envoy:
    """Envoy class."""

    def __init__(
            self, *,
            shard_name: str,
            director_host: str,
            director_port: int,
            shard_descriptor: Type[ShardDescriptor],
            root_certificate: Optional[Union[Path, str]] = None,
            private_key: Optional[Union[Path, str]] = None,
            certificate: Optional[Union[Path, str]] = None,
            tls: bool = True,
            cuda_devices: Union[tuple, list] = (),
            cuda_device_monitor: Optional[Type[CUDADeviceMonitor]] = None,
    ) -> None:
        """Initialize a envoy object."""
        self.name = shard_name
        self.root_certificate = Path(
            root_certificate).absolute() if root_certificate is not None else None
        self.private_key = Path(private_key).absolute() if root_certificate is not None else None
        self.certificate = Path(certificate).absolute() if root_certificate is not None else None
        self.director_client = ShardDirectorClient(
            director_host=director_host,
            director_port=director_port,
            shard_name=shard_name,
            tls=tls,
            root_certificate=root_certificate,
            private_key=private_key,
            certificate=certificate
        )

        self.shard_descriptor = shard_descriptor
        self.cuda_devices = tuple(cuda_devices)

        # Optional plugins
        self.cuda_device_monitor = cuda_device_monitor

        self.executor = ThreadPoolExecutor()
        self.running_experiments = {}
        self.is_experiment_running = False
        self._health_check_future = None

    def run(self):
        """Run of the envoy working cycle."""
        while True:
            try:
                # Workspace import should not be done by gRPC client!
                experiment_name = self.director_client.wait_experiment()
                data_stream = self.director_client.get_experiment_data(experiment_name)
            except Exception as exc:
                logger.exception(f'Failed to get experiment: {exc}')
                time.sleep(DEFAULT_RETRY_TIMEOUT_IN_SECONDS)
                continue
            data_file_path = self._save_data_stream_to_file(data_stream)
            self.is_experiment_running = True
            try:
                with ExperimentWorkspace(
                        self.name + '_' + experiment_name,
                        data_file_path,
                        is_install_requirements=True
                ):
                    self._run_collaborator()
            except Exception as exc:
                logger.exception(f'Collaborator failed with error: {exc}:')
                self.director_client.set_experiment_failed(
                    experiment_name,
                    error_code=1,
                    error_description=traceback.format_exc()
                )
            finally:
                self.is_experiment_running = False

    @staticmethod
    def _save_data_stream_to_file(data_stream):
        data_file_path = Path(str(uuid.uuid4())).absolute()
        with open(data_file_path, 'wb') as data_file:
            for response in data_stream:
                if response.size == len(response.npbytes):
                    data_file.write(response.npbytes)
                else:
                    raise Exception('Broken archive')
        return data_file_path

    def send_health_check(self):
        """Send health check to the director."""
        logger.info('The health check sender is started.')
        while True:
            cuda_devices_info = self._get_cuda_device_info()
            timeout = self.director_client.send_health_check(
                envoy_name=self.name,
                is_experiment_running=self.is_experiment_running,
                cuda_devices_info=cuda_devices_info,
            )
            time.sleep(timeout)

    def _get_cuda_device_info(self):
        cuda_devices_info = None
        try:
            if self.cuda_device_monitor is not None:
                cuda_devices_info = []
                cuda_driver_version = self.cuda_device_monitor.get_driver_version()
                cuda_version = self.cuda_device_monitor.get_cuda_version()
                for device_id in self.cuda_devices:
                    memory_total = self.cuda_device_monitor.get_device_memory_total(device_id)
                    memory_utilized = self.cuda_device_monitor.get_device_memory_utilized(
                        device_id
                    )
                    device_utilization = self.cuda_device_monitor.get_device_utilization(device_id)
                    device_name = self.cuda_device_monitor.get_device_name(device_id)
                    cuda_devices_info.append({
                        'index': device_id,
                        'memory_total': memory_total,
                        'memory_utilized': memory_utilized,
                        'device_utilization': device_utilization,
                        'cuda_driver_version': cuda_driver_version,
                        'cuda_version': cuda_version,
                        'name': device_name,
                    })
        except Exception as exc:
            logger.exception(f'Failed to get cuda device info: {exc}. '
                             f'Check your cuda device monitor plugin.')
        return cuda_devices_info

    def _run_collaborator(self, plan='plan/plan.yaml'):
        """Run the collaborator for the experiment running."""
        plan = Plan.parse(plan_config_path=Path(plan))

        # TODO: Need to restructure data loader config file loader
        echo(f'Data = {plan.cols_data_paths}')
        logger.info('🧿 Starting a Collaborator Service.')

        col = plan.get_collaborator(self.name, self.root_certificate, self.private_key,
                                    self.certificate, shard_descriptor=self.shard_descriptor)
        col.set_available_devices(cuda=self.cuda_devices)
        col.run()

    def start(self):
        """Start the envoy."""
        try:
            is_accepted = self.director_client.report_shard_info(
                shard_descriptor=self.shard_descriptor,
                cuda_devices=self.cuda_devices)
        except Exception as exc:
            logger.exception(f'Failed to report shard info: {exc}')
        else:
            if is_accepted:
                # Shard accepted for participation in the federation
                logger.info('Shard accepted')
                self._health_check_future = self.executor.submit(self.send_health_check)
                self.run()
            else:
                # Shut down
                logger.error('Report shard info was not accepted')



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Aggregator module."""

import sys
from logging import getLogger

from click import echo
from click import group
from click import option
from click import pass_context
from click import Path as ClickPath
from click import style

from openfl.utilities import click_types
from openfl.utilities.path_check import is_directory_traversal
from openfl.utilities.utils import getfqdn_env

logger = getLogger(__name__)


@group()
@pass_context
def aggregator(context):
    """Manage Federated Learning Aggregator."""
    context.obj['group'] = 'aggregator'


@aggregator.command(name='start')
@option('-p', '--plan', required=False,
        help='Federated learning plan [plan/plan.yaml]',
        default='plan/plan.yaml',
        type=ClickPath(exists=True))
@option('-c', '--authorized_cols', required=False,
        help='Authorized collaborator list [plan/cols.yaml]',
        default='plan/cols.yaml', type=ClickPath(exists=True))
@option('-s', '--secure', required=False,
        help='Enable Intel SGX Enclave', is_flag=True, default=False)
def start_(plan, authorized_cols, secure):
    """Start the aggregator service."""
    from pathlib import Path

    from openfl.federated import Plan

    if is_directory_traversal(plan):
        echo('Federated learning plan path is out of the openfl workspace scope.')
        sys.exit(1)
    if is_directory_traversal(authorized_cols):
        echo('Authorized collaborator list file path is out of the openfl workspace scope.')
        sys.exit(1)

    plan = Plan.parse(plan_config_path=Path(plan).absolute(),
                      cols_config_path=Path(authorized_cols).absolute())

    logger.info('🧿 Starting the Aggregator Service.')

    plan.get_server().serve()


@aggregator.command(name='generate-cert-request')
@option('--fqdn', required=False, type=click_types.FQDN,
        help=f'The fully qualified domain name of'
             f' aggregator node [{getfqdn_env()}]',
        default=getfqdn_env())
def _generate_cert_request(fqdn):
    generate_cert_request(fqdn)


def generate_cert_request(fqdn):
    """Create aggregator certificate key pair."""
    from openfl.cryptography.participant import generate_csr
    from openfl.cryptography.io import write_crt
    from openfl.cryptography.io import write_key
    from openfl.interface.cli_helper import CERT_DIR

    if fqdn is None:
        fqdn = getfqdn_env()

    common_name = f'{fqdn}'.lower()
    subject_alternative_name = f'DNS:{common_name}'
    file_name = f'agg_{common_name}'

    echo(f'Creating AGGREGATOR certificate key pair with following settings: '
         f'CN={style(common_name, fg="red")},'
         f' SAN={style(subject_alternative_name, fg="red")}')

    server_private_key, server_csr = generate_csr(common_name, server=True)

    (CERT_DIR / 'server').mkdir(parents=True, exist_ok=True)

    echo('  Writing AGGREGATOR certificate key pair to: ' + style(
        f'{CERT_DIR}/server', fg='green'))

    # Write aggregator csr and key to disk
    write_crt(server_csr, CERT_DIR / 'server' / f'{file_name}.csr')
    write_key(server_private_key, CERT_DIR / 'server' / f'{file_name}.key')


# TODO: function not used
def find_certificate_name(file_name):
    """Search the CRT for the actual aggregator name."""
    # This loop looks for the collaborator name in the key
    with open(file_name, 'r') as f:
        for line in f:
            if 'Subject: CN=' in line:
                col_name = line.split('=')[-1].strip()
                break
    return col_name


@aggregator.command(name='certify')
@option('-n', '--fqdn', type=click_types.FQDN,
        help=f'The fully qualified domain name of aggregator node [{getfqdn_env()}]',
        default=getfqdn_env())
@option('-s', '--silent', help='Do not prompt', is_flag=True)
def _certify(fqdn, silent):
    certify(fqdn, silent)


def certify(fqdn, silent):
    """Sign/certify the aggregator certificate key pair."""
    from pathlib import Path

    from click import confirm

    from openfl.cryptography.ca import sign_certificate
    from openfl.cryptography.io import read_crt
    from openfl.cryptography.io import read_csr
    from openfl.cryptography.io import read_key
    from openfl.cryptography.io import write_crt
    from openfl.interface.cli_helper import CERT_DIR

    if fqdn is None:
        fqdn = getfqdn_env()

    common_name = f'{fqdn}'.lower()
    file_name = f'agg_{common_name}'
    cert_name = f'server/{file_name}'
    signing_key_path = 'ca/signing-ca/private/signing-ca.key'
    signing_crt_path = 'ca/signing-ca.crt'

    # Load CSR
    csr_path_absolute_path = Path(CERT_DIR / f'{cert_name}.csr').absolute()
    if not csr_path_absolute_path.exists():
        echo(style('Aggregator certificate signing request not found.', fg='red')
             + ' Please run `fx aggregator generate-cert-request`'
               ' to generate the certificate request.')

    csr, csr_hash = read_csr(csr_path_absolute_path)

    # Load private signing key
    private_sign_key_absolute_path = Path(CERT_DIR / signing_key_path).absolute()
    if not private_sign_key_absolute_path.exists():
        echo(style('Signing key not found.', fg='red')
             + ' Please run `fx workspace certify`'
               ' to initialize the local certificate authority.')

    signing_key = read_key(private_sign_key_absolute_path)

    # Load signing cert
    signing_crt_absolute_path = Path(CERT_DIR / signing_crt_path).absolute()
    if not signing_crt_absolute_path.exists():
        echo(style('Signing certificate not found.', fg='red')
             + ' Please run `fx workspace certify`'
               ' to initialize the local certificate authority.')

    signing_crt = read_crt(signing_crt_absolute_path)

    echo('The CSR Hash for file '
         + style(f'{cert_name}.csr', fg='green')
         + ' = '
         + style(f'{csr_hash}', fg='red'))

    crt_path_absolute_path = Path(CERT_DIR / f'{cert_name}.crt').absolute()

    if silent:

        echo(' Signing AGGREGATOR certificate')
        signed_agg_cert = sign_certificate(csr, signing_key, signing_crt.subject)
        write_crt(signed_agg_cert, crt_path_absolute_path)

    else:

        if confirm('Do you want to sign this certificate?'):

            echo(' Signing AGGREGATOR certificate')
            signed_agg_cert = sign_certificate(csr, signing_key, signing_crt.subject)
            write_crt(signed_agg_cert, crt_path_absolute_path)

        else:
            echo(style('Not signing certificate.', fg='red')
                 + ' Please check with this AGGREGATOR to get the correct'
                   ' certificate for this federation.')



# Copyright (C) 2020-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Noisy-Sin Shard Descriptor."""

from typing import List

import jax.numpy as jnp
from sklearn.datasets import make_regression
from sklearn.model_selection import train_test_split

from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor

class RegressionShardDescriptor(ShardDescriptor):
    """Regression Shard descriptor class."""

    def __init__(self, rank_worldsize: str = '1, 1', **kwargs) -> None:
        """
        Initialize Regression Data Shard Descriptor.

        This Shard Descriptor generate random regression data with some gaussian centered noise
        using make_regression method from sklearn.datasets.
        Shards data across participants using rank and world size.
        """
        
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))
        X_train, y_train, X_test, y_test = self.generate_data()
        self.data_by_type = {
            'train': jnp.concatenate((X_train, y_train[:, None]), axis=1),
            'val': jnp.concatenate((X_test, y_test[:, None]), axis=1)
        }

    def generate_data(self):
        """Generate regression dataset with predefined params."""
        x, y = make_regression(n_samples=1000, n_features=1, noise=14, random_state=24)
        X_train, X_test, y_train, y_test = train_test_split(x, y, random_state=24)
        self.data = jnp.concatenate((x, y[:, None]), axis=1)
        return X_train, y_train, X_test, y_test

    def get_shard_dataset_types(self) -> List[str]:
        """Get available shard dataset types."""
        return list(self.data_by_type)

    def get_dataset(self, dataset_type='train'):
        """Return a shard dataset by type."""
        if dataset_type not in self.data_by_type:
            raise Exception(f'Incorrect dataset type: {dataset_type}')

        if dataset_type in ['train', 'val']:
            return self.data_by_type[dataset_type][self.rank - 1::self.worldsize]
        else:
            raise ValueError

    @property
    def sample_shape(self) -> List[str]:
        """Return the sample shape info."""
        (*x, _) = self.data[0]
        return [str(i) for i in jnp.array(x, ndmin=1).shape]

    @property
    def target_shape(self) -> List[str]:
        """Return the target shape info."""
        (*_, y) = self.data[0]
        return [str(i) for i in jnp.array(y, ndmin=1).shape]

    @property
    def dataset_description(self) -> str:
        """Return the dataset description."""
        return (f'Regression dataset, shard number {self.rank}'
                f' out of {self.worldsize}')
